"""Utility classes and functions for configuring and setting up the content and the look of a QR code."""
import datetime
import decimal
from collections import namedtuple
from dataclasses import asdict
from datetime import date
from enum import Enum
from typing import Optional, Any, Union, Sequence, List, Tuple

import pytz
from django.utils.html import escape
from pydantic import validate_arguments
from pydantic.dataclasses import dataclass as pydantic_dataclass
from qr_code.qrcode.constants import DEFAULT_MODULE_SIZE, SIZE_DICT, DEFAULT_ERROR_CORRECTION, DEFAULT_IMAGE_FORMAT

from segno import helpers


class QRCodeOptions:
    """
    Represents the options used to create and draw a QR code.
    """

    @validate_arguments
    def __init__(
        self,
        size: Union[int, str, None] = DEFAULT_MODULE_SIZE,
        border: int = 4,
        version: Union[int, str, None] = None,
        image_format: str = "svg",
        error_correction: str = DEFAULT_ERROR_CORRECTION,
        encoding: Optional[str] = "utf-8",
        boost_error: bool = True,
        micro: bool = False,
        eci: bool = False,
        dark_color: Union[tuple, str, bool, None] = "#000",
        light_color: Union[tuple, str, bool, None] = "#fff",
        finder_dark_color: Union[tuple, str, bool, None] = False,
        finder_light_color: Union[tuple, str, bool, None] = False,
        data_dark_color: Union[tuple, str, bool, None] = False,
        data_light_color: Union[tuple, str, bool, None] = False,
        version_dark_color: Union[tuple, str, bool, None] = False,
        version_light_color: Union[tuple, str, bool, None] = False,
        format_dark_color: Union[tuple, str, bool, None] = False,
        format_light_color: Union[tuple, str, bool, None] = False,
        alignment_dark_color: Union[tuple, str, bool, None] = False,
        alignment_light_color: Union[tuple, str, bool, None] = False,
        timing_dark_color: Union[tuple, str, bool, None] = False,
        timing_light_color: Union[tuple, str, bool, None] = False,
        separator_color: Union[tuple, str, bool, None] = False,
        dark_module_color: Union[tuple, str, bool, None] = False,
        quiet_zone_color: Union[tuple, str, bool, None] = False,
    ) -> None:
        """
        :param size: The size of the QR code as an integer or a string. Default is *'m'*.
        :type: str or int
        :param int border: The size of the border (blank space around the code).
        :param version: The version of the QR code gives the size of the matrix.
            Default is *None* which mean automatic in order to avoid data overflow.
        :param version: QR Code version. If the value is ``None`` (default), the
            minimal version which fits for the input data will be used.
            Valid values: "M1", "M2", "M3", "M4" (for Micro QR codes) or an
            integer between 1 and 40 (for QR codes).
            The `version` parameter is case insensitive.
        :type version: int, str or None
        :param str image_format: The graphics format used to render the QR code.
            It can be either *'svg'* or *'png'*. Default is *'svg'*.
        :param str error_correction: How much error correction that might be required
            to read the code. It can be either *'L'*, *'M'*, *'Q'*, or *'H'*. Default is *'M'*.
        :param bool boost_error: Tells whether the QR code encoding engine tries to increase the error correction level
            if it does not affect the version. Error correction level is not increased when it impacts the version of
            the code.
        :param bool micro: Indicates if a Micro QR Code should be created. Default: False
        :param encoding: Indicates the encoding in mode "byte". By default
            `encoding` is ``UTF-8``. When set to ``None``, the implementation tries to use the standard conform
            ISO/IEC 8859-1 encoding and if it does not fit, it will use UTF-8. Note that no ECI mode indicator is
            inserted by default (see :paramref:`eci`).
            The `encoding` parameter is case-insensitive.
        :type encoding: str or None
        :param bool eci: Indicates if binary data which does not use the default
            encoding (ISO/IEC 8859-1) should enforce the ECI mode. Since a lot
            of QR code readers do not support the ECI mode, this feature is
            disabled by default and the data is encoded in the provided
            `encoding` using the usual "byte" mode. Set `eci` to ``True`` if
            an ECI header should be inserted into the QR Code. Note that
            the implementation may not know the ECI designator for the provided
            `encoding` and may raise an exception if the ECI designator cannot
            be found.
            The ECI mode is not supported by Micro QR Codes.
        :param dark_color: Color of the dark modules (default: black). The
            color can be provided as ``(R, G, B)`` tuple, as hexadecimal
            format (``#RGB``, ``#RRGGBB`` ``RRGGBBAA``), or web color
            name (i.e. ``red``). If alpha transparency is supported (i.e. PNG and SVG), hexadecimal values like #RRGGBBAA are accepted.
        :param light_color: Color of the light modules (default: white).
            See `color` for valid values. If light is set to ``None`` the
            light modules will be transparent.
        :param finder_dark_color: Color of the dark finder modules (default: same as ``dark_color``)
        :param finder_light_color: Color of the light finder modules (default: same as ``light_color``)
        :param data_dark_color: Color of the dark data modules (default: same as ``dark_color``)
        :param data_light_color: Color of the light data modules (default: same as ``light_color``)
        :param version_dark_color: Color of the dark version modules (default: same as ``dark_color``)
        :param version_light_color: Color of the light version modules (default: same as ``light_color``)
        :param format_dark_color: Color of the dark format modules (default: same as ``dark_color``)
        :param format_light_color: Color of the light format modules (default: same as ``light_color``)
        :param alignment_dark_color: Color of the dark alignment modules (default: same as ``dark_color``)
        :param alignment_light_color: Color of the light alignment modules (default: same as ``light_color``)
        :param timing_dark_color: Color of the dark timing pattern modules (default: same as ``dark_color``)
        :param timing_light_color: Color of the light timing pattern modules (default: same as ``light_color``)
        :param separator_color: Color of the separator (default: same as ``light_color``)
        :param dark_module_color: Color of the dark module (default: same as ``dark_color``)
        :param quiet_zone_color: Color of the quiet zone modules (default: same as ``light_color``)

        The *size* parameter gives the size of each module of the QR code matrix. It can be either a positive integer or one of the following letters:
            * t or T: tiny (value: 6)
            * s or S: small (value: 12)
            * m or M: medium (value: 18)
            * l or L: large (value: 30)
            * h or H: huge (value: 48)

        For PNG image format the size unit is in pixels, while the unit is 0.1 mm for SVG format.

        The *border* parameter controls how many modules thick the border should be (blank space around the code).
        The default is 4, which is the minimum according to the specs.

        The *version* parameter is an integer from 1 to 40 that controls the size of the QR code matrix. Set to None to
        determine this automatically. The smallest, version 1, is a 21 x 21 matrix. The biggest, version 40, is
        177 x 177 matrix.
        The size grows by 4 modules/side.
        For Micro QR codes, valid values are "M1", "M2", "M3", "M4".

        There are 4 error correction levels used for QR codes, with each one adding different amounts of "backup" data
        depending on how much damage the QR code is expected to suffer in its intended environment, and hence how much
        error correction may be required. The correction level can be configured with the *error_correction* parameter as follow:
            * l or L: error correction level L – up to 7% damage
            * m or M: error correction level M – up to 15% damage
            * q or Q: error correction level Q – up to 25% damage
            * h or H: error correction level H – up to 30% damage

        You may enforce the creation of a Micro QR Code with `micro=True`. The `micro` option defaults to `False`.

        The `encoding` option controls the text encoding used in mode "byte" (used for any general text content). By default `encoding` is ``UTF-8``. When set to ``None``, the implementation (based on Segno) tries to use the standard conform ISO/IEC 8859-1 encoding and if it does not fit, it will use UTF-8. Note that no ECI mode indicator is inserted by default (see `eci` option). The `encoding` parameter is case-insensitive.

        The `boost_error` indicates whether the QR code encoding engine (Segno) tries to increase the error correction level if it does not affect the version. Error correction level is not increased when it impacts the version of the code.

        The `eci` option indicates if binary data which does not use the default encoding (ISO/IEC 8859-1) should enforce the ECI mode. Since a lot of QR code readers do not support the ECI mode, this feature is disabled by default and the data is encoded in the provided encoding using the usual “byte” mode. Set eci to `True` if an ECI header should be inserted into the QR Code. Note that the implementation may not know the ECI designator for the provided encoding and may raise an exception if the ECI designator cannot be found. The ECI mode is not supported by Micro QR Codes.

        :raises: TypeError in case an unknown argument is given.
        """
        self._size = size
        self._border = int(border)
        if _can_be_cast_to_int(version):
            version = int(version)  # type: ignore
            if not 1 <= version <= 40:
                version = None
        elif version in ("m1", "m2", "m3", "m4", "M1", "M2", "M3", "M4"):
            version = version.lower()  # type: ignore
            # Set / change the micro setting otherwise Segno complains about
            # conflicting parameters
            micro = True
        else:
            version = None
        self._version = version
        # if not isinstance(micro, bool):
        #     micro = micro == 'True'
        self._micro = micro
        # if not isinstance(eci, bool):
        #     eci = eci == 'True'
        self._eci = eci
        try:
            error = error_correction.lower()
            self._error_correction = error if error in ("l", "m", "q", "h") else DEFAULT_ERROR_CORRECTION
        except AttributeError:
            self._error_correction = DEFAULT_ERROR_CORRECTION
        self._boost_error = boost_error
        # Handle encoding
        self._encoding = None if encoding == "" else encoding
        try:
            image_format = image_format.lower()
            self._image_format = image_format if image_format in ("svg", "png") else DEFAULT_IMAGE_FORMAT
        except AttributeError:
            self._image_format = DEFAULT_IMAGE_FORMAT
        self._colors = dict(
            dark_color=dark_color,
            light_color=light_color,
            finder_dark_color=finder_dark_color,
            finder_light_color=finder_light_color,
            data_dark_color=data_dark_color,
            data_light_color=data_light_color,
            version_dark_color=version_dark_color,
            version_light_color=version_light_color,
            format_dark_color=format_dark_color,
            format_light_color=format_light_color,
            alignment_dark_color=alignment_dark_color,
            alignment_light_color=alignment_light_color,
            timing_dark_color=timing_dark_color,
            timing_light_color=timing_light_color,
            separator_color=separator_color,
            dark_module_color=dark_module_color,
            quiet_zone_color=quiet_zone_color,
        )

    def kw_make(self):
        """Internal method which returns a dict of parameters to create a QR code.

        :rtype: dict
        """
        return dict(
            version=self._version,
            error=self._error_correction,
            micro=self._micro,
            eci=self._eci,
            boost_error=self._boost_error,
            encoding=self._encoding,
        )

    def kw_save(self):
        """Internal method which returns a dict of parameters to save a QR code.

        :rtype: dict
        """
        image_format = self._image_format
        kw = dict(border=self.border, kind=image_format, scale=self._size_as_int())
        # Change the color mapping into the keywords Segno expects
        # (remove the "_color" suffix from the module names)
        kw.update({k[:-6]: v for k, v in self.color_mapping().items()})
        if image_format == "svg":
            kw["unit"] = "mm"
            scale = decimal.Decimal(kw["scale"]) / 10
            kw["scale"] = scale
        return kw

    def color_mapping(self):
        """Internal method which returns the color mapping.

        Only non-default values are returned.

        :rtype: dict
        """
        colors = {k: v for k, v in self._colors.items() if v is not False}
        # Remove common default "dark" and "light" values
        if colors.get("dark_color") in ("#000", "#000000", "black"):
            del colors["dark_color"]
        if colors.get("light_color") in ("#fff", "#FFF", "#ffffff", "#FFFFFF", "white"):
            del colors["light_color"]
        return colors

    def _size_as_int(self):
        """Returns the size as integer value.

        :rtype: int
        """
        size = self._size
        if _can_be_cast_to_int(size):
            actual_size = int(size)
            if actual_size < 1:
                actual_size = SIZE_DICT[DEFAULT_MODULE_SIZE]
        elif isinstance(size, str):
            actual_size = SIZE_DICT.get(size.lower(), DEFAULT_MODULE_SIZE)
        else:
            actual_size = SIZE_DICT[DEFAULT_MODULE_SIZE]
        return actual_size

    @property
    def size(self):
        return self._size

    @property
    def border(self):
        return self._border

    @property
    def version(self):
        return self._version

    @property
    def image_format(self):
        return self._image_format

    @property
    def error_correction(self):
        return self._error_correction

    @property
    def boost_error(self):
        return self._boost_error

    @property
    def micro(self):
        return self._micro

    @property
    def encoding(self):
        return self._encoding

    @property
    def eci(self):
        return self._eci


def _can_be_cast_to_int(value: Any) -> bool:
    return isinstance(value, int) or (isinstance(value, str) and value.isdigit())


class EventClass(Enum):
    PUBLIC = 1
    PRIVATE = 2
    CONFIDENTIAL = 3


class EventStatus(Enum):
    TENTATIVE = 1
    CONFIRMED = 2
    CANCELLED = 3


class EventTransparency(Enum):
    OPAQUE = 1
    TRANSPARENT = 2


@pydantic_dataclass
class VEvent:
    """
    Data for representing VEVENT for iCalendar (.ics) event.

    Only a subset of https://icalendar.org/iCalendar-RFC-5545/3-6-1-event-component.html is supported.

    Fields meaning:
        * uid: Event identifier
        * summary: This property defines a short summary or subject for the calendar event.
        * start: Start of event
        * end: End of event
        * dtstamp: The property indicates the date/time that the instance of the iCalendar object was created. Defaults to current time in UTC.
        * description: This property provides a more complete description of the calendar component, than that provided by the "SUMMARY" property.
        * organizer: E-mail of organizer
        * status: Status of event
        * location: Location of event
        * geo: This property specifies information related to the global position of event. The property value specifies latitude and longitude, in that order. Whole degrees of latitude shall be represented by a two-digit decimal number ranging from -90 through 90. The longitude represents the location east or west of the prime meridian as a positive or negative real number, respectively (a decimal number ranging from -180 through 180).
        * event_class: Either PUBLIC, PRIVATE or CONFIDENTIAL (see `utils.EventClass`).
        * categories: This property defines the categories for calendar event.
        * transparency: Tell whether the event can have its Time Transparency set to "TRANSPARENT" in order to prevent blocking of the event in searches for busy time.
        * url: This property defines a Uniform Resource Locator (URL) associated with the iCalendar object.
    """

    uid: str
    summary: str
    start: datetime.datetime
    end: datetime.datetime
    dtstamp: Optional[datetime.datetime] = None
    description: Optional[str] = None
    organizer: Optional[str] = None
    status: Optional[EventStatus] = None
    location: Optional[str] = None
    geo: Optional[Tuple[float, float]] = None
    event_class: Optional[EventClass] = None
    categories: Optional[List[str]] = None
    transparency: Optional[EventTransparency] = None
    url: Optional[str] = None

    def make_qr_code_data(self) -> str:
        """\
        Creates a string encoding the event information.

        Only a subset of https://icalendar.org/iCalendar-RFC-5545/3-6-1-event-component.html is supported.
        """

        # Inspired form icalendar: https://github.com/collective/icalendar/
        def fold_icalendar_line(text, limit=75, fold_sep="\r\n "):
            """Make a string folded as defined in RFC5545
            Lines of text SHOULD NOT be longer than 75 octets, excluding the line
            break.  Long content lines SHOULD be split into a multiple line
            representations using a line "folding" technique.  That is, a long
            line can be split between any two characters by inserting a CRLF
            immediately followed by a single linear white-space character (i.e.,
            SPACE or HTAB).
            """
            new_text = ""
            for line in text.split("\n"):
                # Use a fast and simple variant for the common case that line is all ASCII.
                try:
                    line.encode("ascii")
                except (UnicodeEncodeError, UnicodeDecodeError):
                    ret_chars = []
                    byte_count = 0
                    for char in line:
                        char_byte_len = len(char.encode("utf-8"))
                        byte_count += char_byte_len
                        if byte_count >= limit:
                            ret_chars.append(fold_sep)
                            byte_count = char_byte_len
                        ret_chars.append(char)
                    new_text += "".join(ret_chars)
                else:
                    new_text += fold_sep.join(line[i : i + limit - 1] for i in range(0, len(line), limit - 1))
            return new_text

        # Source form icalendar: https://github.com/collective/icalendar/
        def escape_char(text):
            """Format value according to iCalendar TEXT escaping rules."""
            # NOTE: ORDER MATTERS!
            return (
                text.replace(r"\N", "\n")
                .replace("\\", "\\\\")
                .replace(";", r"\;")
                .replace(",", r"\,")
                .replace("\r\n", r"\n")
                .replace("\n", r"\n")
            )

        def is_naive_datetime(t) -> bool:
            return t.tzinfo is None or t.tzinfo.utcoffset(t) is None

        def get_datetime_str(t) -> str:
            if is_naive_datetime(t):
                return t.strftime("%Y%m%dT%H%M%S")
            else:
                t_utc = t.astimezone(pytz.utc)
                return t_utc.strftime("%Y%m%dT%H%M%SZ")

        event_str = f"""BEGIN:VCALENDAR
PRODID:Django QR Code
VERSION:2.0
BEGIN:VEVENT
DTSTAMP:{(self.dtstamp or datetime.datetime.utcnow()).astimezone(pytz.utc).strftime("%Y%m%dT%H%M%SZ")}
UID:{self.uid}
DTSTART:{get_datetime_str(self.start)}
DTEND:{get_datetime_str(self.end)}
SUMMARY:{escape_char(self.summary)}"""
        if self.event_class:
            event_str += f"\nCLASS:{self.event_class.name}"
        if self.categories:
            event_str += "\n" + fold_icalendar_line(f"CATEGORIES:{','.join(map(escape_char, self.categories))}")
        if self.transparency:
            event_str += f"\nTRANSP:{self.transparency.name}"
        if self.description:
            event_str += "\n" + fold_icalendar_line(f"DESCRIPTION:{escape_char(self.description)}")
        if self.organizer:
            event_str += f"\nORGANIZER:MAILTO:{self.organizer}"
        if self.status:
            event_str += f"\nSTATUS:{self.status.name}"
        if self.location:
            event_str += "\n" + fold_icalendar_line(f"LOCATION:{escape_char(self.location)}")
        if self.geo:
            event_str += f"\nGEO:{self.geo[0]};{self.geo[1]}"
        if self.url:
            event_str += f"\nURL:{self.url}"
        event_str += "\nEND:VEVENT\nEND:VCALENDAR"
        # print(event_str)
        return event_str


@pydantic_dataclass
class EpcData:
    """
    Data for representing an European Payments Council Quick Response Code (EPC QR Code) version 002.

    You must always use the error correction level "M" and utilizes max. version 13 to fulfill the constraints of the
        EPC QR Code standard.

        .. note::

            Either the ``text`` or ``reference`` must be provided but not both

        .. note::

            Neither the IBAN, BIC, nor remittance reference number or any other
            information is validated (aside from checks regarding the allowed string
            lengths).

    Fields meaning:
        * name: Name of the recipient.
        * iban: International Bank Account Number (IBAN)
        * amount: The amount to transfer. The currency is always Euro, no other currencies are supported.
        * text: Remittance Information (unstructured)
        * reference: Remittance Information (structured)
        * bic: Bank Identifier Code (BIC). Optional, only required for non-EEA countries.
        * purpose: SEPA purpose code.
    """

    name: str
    iban: str
    amount: Union[int, float, decimal.Decimal]
    text: Optional[str] = None
    reference: Optional[str] = None
    bic: Optional[str] = None
    purpose: Optional[str] = None

    def make_qr_code_data(self) -> str:
        """
        Validates the input and creates the data for an European Payments Council Quick Response Code
        (EPC QR Code) version 002.

        This is a wrapper for :py:func:`segno.helpers._make_epc_qr_data` with no choice for encoding.

        :rtype: str
        """
        return helpers._make_epc_qr_data(**asdict(self), encoding=1)  # type: ignore


class ContactDetail:
    """
    Represents the detail of a contact for MeCARD encoding.

    .. note::
        This is a legacy class. Please use :py:class:`MeCard` instead for new projects.

    Fields meaning:
        * first_name
        * last_name
        * first_name_reading: the sound of the first name.
        * last_name_reading: the sound of the last name.
        * tel: the phone number, it can appear multiple times.
        * tel_av: the video-phone number, it can appear multiple times.
        * email: the email address, it can appear multiple times.
        * memo: notes.
        * birthday: the birth date (Python date).
        * address: the fields divided by commas (,) denote PO box, room number, house number, city, prefecture, zip code and country, in order.
        * url: homepage URL.
        * nickname: display name.
        * org: organization or company name (non-standard,but often recognized, ORG field).
    """

    @validate_arguments
    def __init__(
        self,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        first_name_reading: Optional[str] = None,
        last_name_reading: Optional[str] = None,
        tel: Optional[str] = None,
        tel_av: Optional[str] = None,
        email: Optional[str] = None,
        memo: Optional[str] = None,
        birthday: Optional[date] = None,
        address: Optional[str] = None,
        url: Optional[str] = None,
        nickname: Optional[str] = None,
        org: Optional[str] = None,
    ):
        self.first_name = first_name
        self.last_name = last_name
        self.first_name_reading = first_name_reading
        self.last_name_reading = last_name_reading
        self.tel = tel
        self.tel_av = tel_av
        self.email = email
        self.memo = memo
        self.birthday = birthday
        self.address = address
        self.url = url
        self.nickname = nickname
        self.org = org

    def make_qr_code_data(self) -> str:
        """
        Make a text for configuring a contact in a phone book. The MeCARD format is used, with an optional, non-standard (but often recognized) ORG field.

        See this archive of the format specifications: https://web.archive.org/web/20160304025131/https://www.nttdocomo.co.jp/english/service/developer/make/content/barcode/function/application/addressbook/index.html

        :return: the MeCARD contact description.
        """

        # See this for an archive of the format specifications:
        # https://web.archive.org/web/20160304025131/https://www.nttdocomo.co.jp/english/service/developer/make/content/barcode/function/application/addressbook/index.html
        contact_text = "MECARD:"
        for name_components_pair in (
            ("N:%s;", (_escape_mecard_special_chars(self.last_name), _escape_mecard_special_chars(self.first_name))),
            ("SOUND:%s;", (_escape_mecard_special_chars(self.last_name_reading), _escape_mecard_special_chars(self.first_name_reading))),
        ):
            if name_components_pair[1][0] and name_components_pair[1][1]:
                name = "%s,%s" % name_components_pair[1]
            else:
                name = name_components_pair[1][0] or name_components_pair[1][1] or ""
            if name:
                contact_text += name_components_pair[0] % name
        if self.tel:
            contact_text += "TEL:%s;" % _escape_mecard_special_chars(self.tel)
        if self.tel_av:
            contact_text += "TEL-AV:%s;" % _escape_mecard_special_chars(self.tel_av)
        if self.email:
            contact_text += "EMAIL:%s;" % _escape_mecard_special_chars(self.email)
        if self.memo:
            contact_text += "NOTE:%s;" % _escape_mecard_special_chars(self.memo)
        if self.birthday:
            # Format date to YYMMDD.
            contact_text += "BDAY:%s;" % self.birthday.strftime("%Y%m%d")
        if self.address:
            contact_text += "ADR:%s;" % self.address
        if self.url:
            contact_text += "URL:%s;" % _escape_mecard_special_chars(self.url)
        if self.nickname:
            contact_text += "NICKNAME:%s;" % _escape_mecard_special_chars(self.nickname)
        # Not standard, but recognized by several readers.
        if self.org:
            contact_text += "ORG:%s;" % _escape_mecard_special_chars(self.org)
        contact_text += ";"
        return contact_text

    def escaped_value(self, field_name: str):
        return _escape_mecard_special_chars(getattr(self, field_name))


@pydantic_dataclass
class MeCard:
    """Represents the detail of a contact for MeCARD encoding.

    Fields meaning:
        * name: Name. If it contains a comma, the first part is treated as lastname and the second part is treated as forename.
        * reading: Designates a text string to be set as the kana name in the phonebook
        * email: E-mail address. Multiple values are allowed.
        * phone: Phone number. Multiple values are allowed.
        * videophone: Phone number for video calls. Multiple values are allowed.
        * memo: A notice for the contact.
        * nickname: Nickname.
        * birthday: Birthday. If a string is provided, it should encode the date as YYYYMMDD value.
        * url: Homepage. Multiple values are allowed.
        * pobox: P.O. box (address information).
        * roomno: Room number (address information).
        * houseno: House number (address information).
        * city: City (address information).
        * prefecture: Prefecture (address information).
        * zipcode: Zip code (address information).
        * country: Country (address information).
        * org: organization or company name (ORG field, non-standard,but often recognized by readers).
    """

    name: str
    reading: Optional[str] = None
    email: Union[str, Sequence[str], None] = None
    phone: Union[str, Sequence[str], None] = None
    videophone: Union[str, Sequence[str], None] = None
    memo: Optional[str] = None
    nickname: Optional[str] = None
    birthday: Union[str, datetime.date, None] = None
    url: Union[str, Sequence[str], None] = None
    pobox: Optional[str] = None
    roomno: Optional[str] = None
    houseno: Optional[str] = None
    city: Optional[str] = None
    prefecture: Optional[str] = None
    zipcode: Union[int, str, None] = None
    country: Optional[str] = None
    org: Optional[str] = None

    def make_qr_code_data(self) -> str:
        """\
        Creates a string encoding the contact information as MeCARD.

        :rtype: str
        """
        kw = asdict(self)
        if self.zipcode is not None and self.zipcode != "":
            kw["zipcode"] = str(self.zipcode)
        org = kw.pop("org")
        contact_text = helpers.make_mecard_data(**kw)
        # Not standard, but recognized by several readers.
        if org:
            contact_text += f"ORG:{_escape_mecard_special_chars(org)};"
        return contact_text


@pydantic_dataclass
class VCard:
    """Represents the detail of a contact for vCard encoding.

    Creates a QR code which encodes a `vCard <https://en.wikipedia.org/wiki/VCard>`_
    version 3.0.

    Only a subset of available `vCard 3.0 properties <https://tools.ietf.org/html/rfc2426>`
    is supported.

    Fields meaning:
    name: The name. If it contains a semicolon, , the first part
            is treated as lastname and the second part is treated as forename.
    displayname: Common name.
    email: E-mail address. Multiple values are allowed.
    phone: Phone number. Multiple values are allowed.
    fax: Fax number. Multiple values are allowed.
    videophone: Phone number for video calls. Multiple values are allowed.
    memo: A notice for the contact.
    nickname: Nickname.
    birthday: Birthday. If a string is provided, it should encode the
                     date as ``YYYY-MM-DD`` value.
    url: Homepage. Multiple values are allowed.
    pobox: P.O. box (address information).
    street: Street address.
    city: City (address information).
    region: Region (address information).
    zipcode: Zip code (address information).
    country: Country (address information).
    org: Company / organization name.
    lat: Latitude.
    lng: Longitude.
    source: URL where to obtain the vCard.
    rev: Revision of the vCard / last modification date.
    title: Job Title. Multiple values are allowed.
    photo_uri: Photo URI. Multiple values are allowed.
    cellphone: Cell phone number. Multiple values are allowed.
    homephone: Home phone number. Multiple values are allowed.
    workphone: Work phone number. Multiple values are allowed.
    """

    name: str
    displayname: Optional[str] = None
    email: Union[str, Sequence[str], None] = None
    phone: Union[str, Sequence[str], None] = None
    fax: Union[str, Sequence[str], None] = None
    videophone: Union[str, Sequence[str], None] = None
    memo: Optional[str] = None
    nickname: Optional[str] = None
    birthday: Union[str, datetime.date, None] = None
    url: Union[str, Sequence[str], None] = None
    pobox: Optional[str] = None
    street: Optional[str] = None
    city: Optional[str] = None
    region: Optional[str] = None
    zipcode: Union[int, str, None] = None
    country: Optional[str] = None
    org: Optional[str] = None
    lat: Optional[float] = None
    lng: Optional[float] = None
    source: Optional[str] = None
    rev: Union[str, datetime.date, None] = None
    title: Union[str, Sequence[str], None] = None
    photo_uri: Union[str, Sequence[str], None] = None
    cellphone: Union[str, Sequence[str], None] = None
    homephone: Union[str, Sequence[str], None] = None
    workphone: Union[str, Sequence[str], None] = None

    def __post_init__(self):
        if self.displayname is None:
            self.displayname = self.name.replace(" ; ", " ").replace("; ", " ").replace(";", " ")

    def make_qr_code_data(self) -> str:
        """\
        Creates a string encoding the contact information as vCard 3.0.

        Only a subset of available `vCard 3.0 properties <https://tools.ietf.org/html/rfc2426>`
        is supported.

        :rtype: str
        """
        kw = asdict(self)
        kw["zipcode"] = str(self.zipcode)
        return helpers.make_vcard_data(**kw)


@pydantic_dataclass
class WifiConfig:
    """\
    Represents a WIFI configuration.

    Fields meaning:
        * ssid: the name of the SSID
        * authentication: the authentication type for the SSID; can be AUTHENTICATION.wep or AUTHENTICATION.wpa, or AUTHENTICATION.nopass for no password. Or, omit for no password.
        * password: the password, ignored if "authentication" is 'nopass' (in which case it may be omitted).
        * hidden: tells whether the SSID is hidden or not; can be True or False.
    """

    AUTHENTICATION = namedtuple("AUTHENTICATION", "nopass WEP WPA")._make(range(3))  # type: ignore
    AUTHENTICATION_CHOICES = ((AUTHENTICATION.nopass, "nopass"), (AUTHENTICATION.WEP, "WEP"), (AUTHENTICATION.WPA, "WPA"))

    def __init__(self, ssid: str = "", authentication: int = AUTHENTICATION.nopass, password: str = "", hidden: bool = False) -> None:
        self.ssid = ssid
        self.authentication = authentication
        self.password = password
        self.hidden = hidden

    def make_qr_code_data(self) -> str:
        """
        Make a text for configuring a Wi-Fi connexion. The syntax is inspired by the MeCARD format used for contacts.

        :return: the WIFI configuration text that can be translated to a QR code.
        :rtype: str
        """

        wifi_config = "WIFI:"
        if self.ssid:
            wifi_config += "S:%s;" % _escape_mecard_special_chars(self.ssid)
        if self.authentication:
            wifi_config += "T:%s;" % WifiConfig.AUTHENTICATION_CHOICES[self.authentication][1]
        if self.password:
            wifi_config += "P:%s;" % _escape_mecard_special_chars(self.password)
        if self.hidden:
            wifi_config += "H:%s;" % str(self.hidden).lower()
        wifi_config += ";"
        return wifi_config


@pydantic_dataclass
class Coordinates:
    """\
    Represents a set of coordinates with an optional altitude.

    Fields meaning:
        * latitude: The latitude.
        * latitude: The longitude.
        * latitude: The optional altitude.
    """

    latitude: float
    longitude: float
    altitude: Optional[Union[int, float]] = None

    def __str__(self) -> str:
        if self.altitude:
            return f"latitude: {self.latitude}, longitude: {self.longitude}, altitude: {self.altitude}"
        return f"latitude: {self.latitude}, longitude: {self.longitude}"

    def float_to_str(self, f):
        return f"{f:.8f}".rstrip("0")

    def make_geolocation_text(self) -> str:
        geo = f"geo:{self.float_to_str(self.latitude)},{self.float_to_str(self.longitude)}"
        if self.altitude:
            return f"{geo},{self.float_to_str(self.altitude)}"
        return geo

    def make_google_maps_text(self) -> str:
        geo = f"https://maps.google.com/local?q={self.float_to_str(self.latitude)},{self.float_to_str(self.longitude)}"
        if self.altitude:
            return f"{geo},{self.float_to_str(self.altitude)}"
        return geo


def make_tel_text(phone_number: Any) -> str:
    return "tel:%s" % phone_number


def make_sms_text(phone_number: Any) -> str:
    return "sms:%s" % phone_number


def make_youtube_text(video_id: str) -> str:
    return f"https://www.youtube.com/watch/?v={escape(video_id)}"


def make_google_play_text(package_id: str) -> str:
    return f"https://play.google.com/store/apps/details?id={escape(package_id)}"


@pydantic_dataclass
class Email:
    """Represents the data of an e-mail.

    Fields meaning:
        * to: The email address (recipient). Multiple values are allowed.
        * cc: The carbon copy recipient. Multiple values are allowed.
        * bcc: The blind carbon copy recipient. Multiple values are allowed.
        * subject: The subject.
        * body: The message body.
    """

    to: Union[str, Sequence[str]]
    cc: Union[str, Sequence[str], None] = None
    bcc: Union[str, Sequence[str], None] = None
    subject: Optional[str] = None
    body: Optional[str] = None

    def make_qr_code_data(self) -> str:
        """\
        Creates either a simple "mailto:" URL or complete e-mail message with
        (blind) carbon copies and a subject and a body.

        :rtype: str
        """
        return helpers.make_make_email_data(**asdict(self))


@validate_arguments
def _escape_mecard_special_chars(string_to_escape: Optional[str]) -> Optional[str]:
    if not string_to_escape:
        return string_to_escape
    special_chars = ["\\", '"', ";", ",", ":"]
    for sc in special_chars:
        string_to_escape = string_to_escape.replace(sc, "\\%s" % sc)
    return string_to_escape



import os
import re

from qr_code.tests import PNG_REF_SUFFIX, SVG_REF_SUFFIX


def get_urls_without_token_for_comparison(*urls):
    token_regex = re.compile(r"&?token=[^&]+")
    simplified_urls = list(map(lambda x: token_regex.sub("", x), urls))
    simplified_urls = list(map(lambda x: x.replace("?&", "?"), simplified_urls))
    return simplified_urls


def get_resources_path():
    tests_dir = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(tests_dir, "resources")


def minimal_svg(s):
    """Returns the SVG document without the XML declaration and SVG namespace declaration

    :rtype: str
    """
    return s.replace('<?xml version="1.0" encoding="utf-8"?>\n', "").replace('xmlns="http://www.w3.org/2000/svg" ', "").strip()


def get_svg_content_from_file_name(base_file_name):
    with open(os.path.join(get_resources_path(), base_file_name + SVG_REF_SUFFIX), encoding="utf-8") as file:
        return file.read()


def get_png_content_from_file_name(base_file_name):
    with open(os.path.join(get_resources_path(), base_file_name + PNG_REF_SUFFIX), "rb") as file:
        return file.read()


def write_svg_content_to_file(base_file_name, image_content):
    with open(os.path.join(get_resources_path(), base_file_name + SVG_REF_SUFFIX), "wt", encoding="utf-8") as file:
        file.write(image_content)


def write_png_content_to_file(base_file_name, image_content):
    with open(os.path.join(get_resources_path(), base_file_name + PNG_REF_SUFFIX), "wb") as file:
        file.write(image_content)



from django.conf import settings
from django.urls import path

from qr_code import views


app_name = "qr_code"
urlpatterns = [
    path(
        getattr(settings, "SERVE_QR_CODE_IMAGE_PATH", "images/serve-qr-code-image/"), views.serve_qr_code_image, name="serve_qr_code_image"
    )
]



"""Tests for qr_code application."""
import base64
import re

from itertools import product

from django.conf import settings
from django.contrib.auth.models import AnonymousUser, User
from django.core.cache import caches
from django.template import Template, Context
from django.test import SimpleTestCase, override_settings
from django.utils.safestring import mark_safe
from django.utils.html import escape

from qr_code.qrcode.maker import make_embedded_qr_code
from qr_code.qrcode.constants import ERROR_CORRECTION_DICT
from qr_code.qrcode.serve import make_qr_code_url, allows_external_request_from_user
from qr_code.qrcode.utils import QRCodeOptions
from qr_code.templatetags.qr_code import qr_from_data, qr_url_from_data

from qr_code.tests import (
    REFRESH_REFERENCE_IMAGES,
    TEST_TEXT,
    OVERRIDE_CACHES_SETTING,
    COMPLEX_TEST_TEXT,
    BASE64_PNG_IMAGE_TEMPLATE,
    IMAGE_TAG_BASE64_DATA_RE,
)
from qr_code.tests.utils import (
    write_svg_content_to_file,
    write_png_content_to_file,
    get_svg_content_from_file_name,
    get_png_content_from_file_name,
    get_urls_without_token_for_comparison,
    minimal_svg,
)

COMPLEX_TEST_TEXT_AS_ISO_8859_1 = COMPLEX_TEST_TEXT.encode("iso-8859-1")
COMPLEX_TEST_TEXT_AS_UTF_8 = COMPLEX_TEST_TEXT.encode("utf-8")
TEST_TEXT_AS_UTF_8 = TEST_TEXT.encode("utf-8")


class TestQRUrlFromDataResult(SimpleTestCase):
    """
    Ensures that serving images representing QR codes works as expected (with or without caching, and with or without
    protection against external requests).
    """

    default_ref_base_file_name = "qrfromdata_default"

    @staticmethod
    def _get_reference_result_for_default_svg():
        return get_svg_content_from_file_name(TestQRUrlFromDataResult.default_ref_base_file_name)

    @staticmethod
    def _get_reference_result_for_default_png():
        return get_png_content_from_file_name(TestQRUrlFromDataResult.default_ref_base_file_name)

    def test_svg_url(self):
        is_first = True
        users = [None, AnonymousUser(), User(username="test")]
        for url_options in product([True, False, None], [True, False, None], users):
            cache_enabled = url_options[0]
            url_signature_enabled = url_options[1]
            user = url_options[2]
            print(f"\t - cache_enabled={cache_enabled}, url_signature_enabled={url_signature_enabled}, user={user}")
            url_options_kwargs = dict()
            url0 = make_qr_code_url(
                TEST_TEXT_AS_UTF_8,
                QRCodeOptions(size=1),
                **url_options_kwargs,
                force_text=False,
                cache_enabled=cache_enabled,
                url_signature_enabled=url_signature_enabled,
            )
            if cache_enabled is not None:
                url_options_kwargs["cache_enabled"] = cache_enabled
            url1 = make_qr_code_url(
                TEST_TEXT_AS_UTF_8,
                QRCodeOptions(size=1),
                **url_options_kwargs,
                force_text=False,
                url_signature_enabled=url_signature_enabled,
            )
            if url_signature_enabled is not None:
                url_options_kwargs["url_signature_enabled"] = url_signature_enabled
            url2 = qr_url_from_data(TEST_TEXT_AS_UTF_8, size=1, **url_options_kwargs)
            url3 = qr_url_from_data(TEST_TEXT_AS_UTF_8, image_format="svg", size=1, **url_options_kwargs)
            url4 = qr_url_from_data(TEST_TEXT_AS_UTF_8, image_format="SVG", size=1, **url_options_kwargs)
            url5 = qr_url_from_data(TEST_TEXT_AS_UTF_8, options=QRCodeOptions(image_format="SVG", size=1), **url_options_kwargs)
            # Using an invalid image format should fallback to SVG.
            url6 = qr_url_from_data(TEST_TEXT_AS_UTF_8, image_format="invalid-format-name", size=1, **url_options_kwargs)
            url = url1
            if url_signature_enabled is not False:
                urls = get_urls_without_token_for_comparison(url0, url1, url2, url3, url4, url5, url6)
            else:
                urls = [url0, url1, url2, url3, url4, url5, url6]
            self.assertEqual(urls[0], urls[1])
            self.assertEqual(urls[0], urls[2])
            self.assertEqual(urls[0], urls[3])
            self.assertEqual(urls[0], urls[4])
            self.assertEqual(urls[0], urls[5])
            response = self.client.get(url)
            expected_status_code = 200
            if url_signature_enabled is False and not allows_external_request_from_user(user):
                expected_status_code = 403
            self.assertEqual(response.status_code, expected_status_code)
            image_data = response.content.decode("utf-8")
            if expected_status_code == 200:
                if is_first and REFRESH_REFERENCE_IMAGES:
                    write_svg_content_to_file(TestQRUrlFromDataResult.default_ref_base_file_name, image_data)
                    is_first = False
                self.assertEqual(image_data, TestQRUrlFromDataResult._get_reference_result_for_default_svg())

    def test_png_url(self):
        is_first = True
        for url_options in product([True, False, None], [True, False, None]):
            cache_enabled = url_options[0]
            url_signature_enabled = url_options[1]
            url_options_kwargs = dict()
            if cache_enabled is not None:
                url_options_kwargs["cache_enabled"] = cache_enabled
            if url_signature_enabled is not None:
                url_options_kwargs["url_signature_enabled"] = url_signature_enabled
            url1 = make_qr_code_url(TEST_TEXT_AS_UTF_8, QRCodeOptions(image_format="png", size=1), **url_options_kwargs, force_text=False)
            url2 = qr_url_from_data(TEST_TEXT_AS_UTF_8, image_format="png", size=1, **url_options_kwargs)
            url3 = qr_url_from_data(TEST_TEXT_AS_UTF_8, image_format="PNG", size=1, **url_options_kwargs)
            url4 = qr_url_from_data(TEST_TEXT_AS_UTF_8, options=QRCodeOptions(image_format="PNG", size=1), **url_options_kwargs)
            url = url1
            if url_signature_enabled is not False:
                urls = get_urls_without_token_for_comparison(url1, url2, url3, url4)
            else:
                urls = [url1, url2, url3, url4]
            self.assertEqual(urls[0], urls[1])
            self.assertEqual(urls[0], urls[2])
            self.assertEqual(urls[0], urls[3])
            response = self.client.get(url)
            print(f"\t - cache_enabled={cache_enabled}, url_signature_enabled={url_signature_enabled}")
            expected_status_code = 200
            if url_signature_enabled is False and not allows_external_request_from_user(None):
                expected_status_code = 403
            self.assertEqual(response.status_code, expected_status_code)
            if expected_status_code == 200:
                if is_first and REFRESH_REFERENCE_IMAGES:
                    write_png_content_to_file(TestQRUrlFromDataResult.default_ref_base_file_name, response.content)
                    is_first = False
                self.assertEqual(response.content, TestQRUrlFromDataResult._get_reference_result_for_default_png())

    @override_settings(CACHES=OVERRIDE_CACHES_SETTING, QR_CODE_CACHE_ALIAS=None)
    def test_svg_with_cache_but_no_alias(self):
        self.test_svg_url()

    @override_settings(CACHES=OVERRIDE_CACHES_SETTING)
    def test_png_with_cache(self):
        self.test_png_url()

    @override_settings(CACHES=OVERRIDE_CACHES_SETTING, QR_CODE_CACHE_ALIAS=None)
    def test_png_with_cache_but_no_alias(self):
        self.test_png_url()

    @override_settings(
        QR_CODE_URL_PROTECTION=dict(
            TOKEN_LENGTH=30,
            SIGNING_KEY="my-secret-signing-key",
            SIGNING_SALT="my-signing-salt",
            ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER=True,
        )
    )
    def test_url_with_protection_settings_1(self):
        # We need to clear cache every time we change the QR_CODE_URL_PROTECTION to avoid incidence between tests.
        caches[settings.QR_CODE_CACHE_ALIAS].clear()
        self.test_svg_url()
        self.test_png_url()
        response = self.client.get(make_qr_code_url(TEST_TEXT_AS_UTF_8, force_text=False, url_signature_enabled=False, cache_enabled=False))
        # Registered users can access the URL externally, but since we are not logged in, we must expect an HTTP 403.
        self.assertEqual(response.status_code, 403)

    @override_settings(QR_CODE_URL_PROTECTION=dict(ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER=False))
    def test_url_with_protection_settings_2(self):
        # We need to clear cache every time we change the QR_CODE_URL_PROTECTION to avoid incidence between tests.
        caches[settings.QR_CODE_CACHE_ALIAS].clear()
        self.test_svg_url()
        self.test_png_url()
        response = self.client.get(make_qr_code_url(TEST_TEXT_AS_UTF_8, force_text=False, url_signature_enabled=False, cache_enabled=False))
        self.assertEqual(response.status_code, 403)
        response = self.client.get(make_qr_code_url(TEST_TEXT_AS_UTF_8, force_text=False, url_signature_enabled=True, cache_enabled=False))
        self.assertEqual(response.status_code, 200)

    @override_settings(QR_CODE_URL_PROTECTION=dict(ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER=lambda user: False))
    def test_url_with_protection_settings_3(self):
        # We need to clear cache every time we change the QR_CODE_URL_PROTECTION to avoid incidence between tests.
        caches[settings.QR_CODE_CACHE_ALIAS].clear()
        self.test_svg_url()
        self.test_png_url()
        response = self.client.get(make_qr_code_url(TEST_TEXT_AS_UTF_8, force_text=False, url_signature_enabled=False, cache_enabled=False))
        self.assertEqual(response.status_code, 403)
        response = self.client.get(make_qr_code_url(TEST_TEXT_AS_UTF_8, force_text=False, url_signature_enabled=True, cache_enabled=False))
        self.assertEqual(response.status_code, 200)

    @override_settings(QR_CODE_URL_PROTECTION=dict(ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER=lambda user: True))
    def test_url_with_protection_settings_4(self):
        # We need to clear cache every time we change the QR_CODE_URL_PROTECTION to avoid incidence between tests.
        caches[settings.QR_CODE_CACHE_ALIAS].clear()
        self.test_svg_url()
        self.test_png_url()
        # The callable for ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER always return True, even for anonymous user.
        # Therefore, we must expect an HTTP 200.
        # We test with different values of url_signature_enabled.
        response = self.client.get(make_qr_code_url(TEST_TEXT_AS_UTF_8, force_text=False, cache_enabled=False))
        self.assertEqual(response.status_code, 200)
        response = self.client.get(make_qr_code_url(TEST_TEXT_AS_UTF_8, force_text=False, url_signature_enabled=True, cache_enabled=False))
        self.assertEqual(response.status_code, 200)
        response = self.client.get(make_qr_code_url(TEST_TEXT_AS_UTF_8, force_text=False, url_signature_enabled=False, cache_enabled=False))
        self.assertEqual(response.status_code, 200)

    def test_url_with_invalid_signature_token(self):
        valid_url_with_signature_token = make_qr_code_url(TEST_TEXT_AS_UTF_8, force_text=False)
        url_with_invalid_signature_token = valid_url_with_signature_token.replace("token=", "token=some-front-padding")
        response = self.client.get(url_with_invalid_signature_token)
        self.assertEqual(response.status_code, 403)

    def test_url_with_wrong_signature_token(self):
        valid_url_with_signature_token_for_size_10 = make_qr_code_url(TEST_TEXT_AS_UTF_8, QRCodeOptions(size=10), force_text=False)
        valid_url_with_signature_token_for_size_8 = make_qr_code_url(TEST_TEXT_AS_UTF_8, QRCodeOptions(size=8), force_text=False)
        token_regex = re.compile(r"token=([^&]+)")
        match = token_regex.search(valid_url_with_signature_token_for_size_8)
        size_8_token_value = match.group(1)
        match = token_regex.search(valid_url_with_signature_token_for_size_10)
        size_10_token_value = match.group(1)
        url_with_invalid_signature_token = valid_url_with_signature_token_for_size_10.replace(size_10_token_value, size_8_token_value)
        response = self.client.get(url_with_invalid_signature_token)
        self.assertEqual(response.status_code, 403)

    def test_svg_error_correction(self):
        base_file_name = "qrfromdata_error_correction"
        for correction_level in ERROR_CORRECTION_DICT:
            print("Testing SVG URL with error correction: %s" % correction_level)
            url1 = make_qr_code_url(
                COMPLEX_TEST_TEXT_AS_UTF_8, QRCodeOptions(error_correction=correction_level), force_text=False, cache_enabled=False
            )
            url2 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, error_correction=correction_level, cache_enabled=False)
            url3 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, error_correction=correction_level, image_format="svg", cache_enabled=False)
            url4 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, error_correction=correction_level, image_format="SVG", cache_enabled=False)
            url5 = qr_url_from_data(
                COMPLEX_TEST_TEXT_AS_UTF_8,
                options=QRCodeOptions(error_correction=correction_level, image_format="SVG"),
                cache_enabled=False,
            )
            # Using an invalid image format should fallback to SVG.
            url6 = qr_url_from_data(
                COMPLEX_TEST_TEXT_AS_UTF_8, error_correction=correction_level, image_format="invalid-format-name", cache_enabled=False
            )
            url = url1
            urls = get_urls_without_token_for_comparison(url1, url2, url3, url4, url5, url6)
            self.assertEqual(urls[0], urls[1])
            self.assertEqual(urls[0], urls[2])
            self.assertEqual(urls[0], urls[3])
            self.assertEqual(urls[0], urls[4])
            self.assertEqual(urls[0], urls[5])
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            source_image_data = response.content.decode("utf-8")
            ref_file_name = f"{base_file_name}_{correction_level.lower()}"
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(ref_file_name, source_image_data)
            ref_image_data = get_svg_content_from_file_name(ref_file_name)
            self.assertEqual(source_image_data, ref_image_data)

    def test_png_error_correction(self):
        base_file_name = "qrfromdata_error_correction"
        for correction_level in ERROR_CORRECTION_DICT:
            print("Testing PNG URL with error correction: %s" % correction_level)
            url1 = make_qr_code_url(
                COMPLEX_TEST_TEXT_AS_UTF_8,
                QRCodeOptions(error_correction=correction_level, image_format="png"),
                force_text=False,
                cache_enabled=False,
            )
            url2 = make_qr_code_url(
                COMPLEX_TEST_TEXT_AS_UTF_8,
                QRCodeOptions(error_correction=correction_level, image_format="PNG"),
                force_text=False,
                cache_enabled=False,
            )
            url3 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, error_correction=correction_level, image_format="png", cache_enabled=False)
            url4 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, error_correction=correction_level, image_format="PNG", cache_enabled=False)
            url5 = qr_url_from_data(
                COMPLEX_TEST_TEXT_AS_UTF_8,
                options=QRCodeOptions(error_correction=correction_level, image_format="PNG"),
                cache_enabled=False,
            )
            url = url1
            urls = get_urls_without_token_for_comparison(url1, url2, url3, url4, url5)
            self.assertEqual(urls[0], urls[1])
            self.assertEqual(urls[0], urls[2])
            self.assertEqual(urls[0], urls[3])
            self.assertEqual(urls[0], urls[4])
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            source_image_data = response.content
            ref_file_name = f"{base_file_name}_{correction_level.lower()}"
            if REFRESH_REFERENCE_IMAGES:
                write_png_content_to_file(ref_file_name, source_image_data)
            ref_image_data = get_png_content_from_file_name(ref_file_name)
            self.assertEqual(source_image_data, ref_image_data)

    def test_svg_eci(self):
        base_file_name = "qrfromdata_eci"
        for eci in [False, True]:
            print("Testing SVG URL with ECI: %s" % eci)
            url1 = make_qr_code_url(COMPLEX_TEST_TEXT_AS_UTF_8, QRCodeOptions(eci=eci), force_text=False, cache_enabled=False)
            url2 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, eci=eci, cache_enabled=False)
            url3 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, eci=eci, image_format="svg", cache_enabled=False)
            url4 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, eci=eci, image_format="SVG", cache_enabled=False)
            url5 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, options=QRCodeOptions(eci=eci, image_format="SVG"), cache_enabled=False)
            # Using an invalid image format should fallback to SVG.
            url6 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, eci=eci, image_format="invalid-format-name", cache_enabled=False)
            url = url1
            urls = get_urls_without_token_for_comparison(url1, url2, url3, url4, url5, url6)
            self.assertEqual(urls[0], urls[1])
            self.assertEqual(urls[0], urls[2])
            self.assertEqual(urls[0], urls[3])
            self.assertEqual(urls[0], urls[4])
            self.assertEqual(urls[0], urls[5])
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            source_image_data = response.content.decode("utf-8")
            ref_file_name = f"{base_file_name}_{str(eci).lower()}"
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(ref_file_name, source_image_data)
            ref_image_data = get_svg_content_from_file_name(ref_file_name)
            self.assertEqual(source_image_data, ref_image_data)

    def test_png_eci(self):
        base_file_name = "qrfromdata_eci"
        for eci in [False, True]:
            print("Testing PNG URL with ECI: %s" % eci)
            url1 = make_qr_code_url(
                COMPLEX_TEST_TEXT_AS_UTF_8, QRCodeOptions(eci=eci, image_format="png"), force_text=False, cache_enabled=False
            )
            url2 = make_qr_code_url(
                COMPLEX_TEST_TEXT_AS_UTF_8, QRCodeOptions(eci=eci, image_format="PNG"), force_text=False, cache_enabled=False
            )
            url3 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, eci=eci, image_format="png", cache_enabled=False)
            url4 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, eci=eci, image_format="PNG", cache_enabled=False)
            url5 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, options=QRCodeOptions(eci=eci, image_format="PNG"), cache_enabled=False)
            url = url1
            urls = get_urls_without_token_for_comparison(url1, url2, url3, url4, url5)
            self.assertEqual(urls[0], urls[1])
            self.assertEqual(urls[0], urls[2])
            self.assertEqual(urls[0], urls[3])
            self.assertEqual(urls[0], urls[4])
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            source_image_data = response.content
            ref_file_name = f"{base_file_name}_{str(eci).lower()}"
            if REFRESH_REFERENCE_IMAGES:
                write_png_content_to_file(ref_file_name, source_image_data)
            ref_image_data = get_png_content_from_file_name(ref_file_name)
            self.assertEqual(source_image_data, ref_image_data)

    def test_svg_micro(self):
        base_file_name = "qrfromdata_micro"
        for micro in [False, True]:
            print("Testing SVG URL with micro: %s" % micro)
            url1 = make_qr_code_url(COMPLEX_TEST_TEXT_AS_ISO_8859_1, QRCodeOptions(micro=micro), force_text=False, cache_enabled=False)
            url2 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_ISO_8859_1, micro=micro, cache_enabled=False)
            url3 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_ISO_8859_1, micro=micro, image_format="svg", cache_enabled=False)
            url4 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_ISO_8859_1, micro=micro, image_format="SVG", cache_enabled=False)
            url5 = qr_url_from_data(
                COMPLEX_TEST_TEXT_AS_ISO_8859_1, options=QRCodeOptions(micro=micro, image_format="SVG"), cache_enabled=False
            )
            # Using an invalid image format should fallback to SVG.
            url6 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_ISO_8859_1, micro=micro, image_format="invalid-format-name", cache_enabled=False)
            url = url1
            urls = get_urls_without_token_for_comparison(url1, url2, url3, url4, url5, url6)
            self.assertEqual(urls[0], urls[1])
            self.assertEqual(urls[0], urls[2])
            self.assertEqual(urls[0], urls[3])
            self.assertEqual(urls[0], urls[4])
            self.assertEqual(urls[0], urls[5])
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            source_image_data = response.content.decode("utf-8")
            ref_file_name = f"{base_file_name}_{str(micro).lower()}"
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(ref_file_name, source_image_data)
            ref_image_data = get_svg_content_from_file_name(ref_file_name)
            self.assertEqual(source_image_data, ref_image_data)

    def test_png_micro(self):
        base_file_name = "qrfromdata_micro"
        for micro in [False, True]:
            print("Testing PNG URL with micro: %s" % micro)
            url1 = make_qr_code_url(
                COMPLEX_TEST_TEXT_AS_ISO_8859_1, QRCodeOptions(micro=micro, image_format="png"), force_text=False, cache_enabled=False
            )
            url2 = make_qr_code_url(
                COMPLEX_TEST_TEXT_AS_ISO_8859_1, QRCodeOptions(micro=micro, image_format="PNG"), force_text=False, cache_enabled=False
            )
            url3 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_ISO_8859_1, micro=micro, image_format="png", cache_enabled=False)
            url4 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_ISO_8859_1, micro=micro, image_format="PNG", cache_enabled=False)
            url5 = qr_url_from_data(
                COMPLEX_TEST_TEXT_AS_ISO_8859_1, options=QRCodeOptions(micro=micro, image_format="PNG"), cache_enabled=False
            )
            url = url1
            urls = get_urls_without_token_for_comparison(url1, url2, url3, url4, url5)
            self.assertEqual(urls[0], urls[1])
            self.assertEqual(urls[0], urls[2])
            self.assertEqual(urls[0], urls[3])
            self.assertEqual(urls[0], urls[4])
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            source_image_data = response.content
            ref_file_name = f"{base_file_name}_{str(micro).lower()}"
            if REFRESH_REFERENCE_IMAGES:
                write_png_content_to_file(ref_file_name, source_image_data)
            ref_image_data = get_png_content_from_file_name(ref_file_name)
            self.assertEqual(source_image_data, ref_image_data)

    def test_svg_boost_error(self):
        base_file_name = "qrfromdata_boost_error"
        for boost_error in [False, True]:
            print("Testing SVG URL with boost_error: %s" % boost_error)
            url1 = make_qr_code_url(
                COMPLEX_TEST_TEXT_AS_UTF_8, QRCodeOptions(boost_error=boost_error), force_text=False, cache_enabled=False
            )
            url2 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, boost_error=boost_error, cache_enabled=False)
            url3 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, boost_error=boost_error, image_format="svg", cache_enabled=False)
            url4 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, boost_error=boost_error, image_format="SVG", cache_enabled=False)
            url5 = qr_url_from_data(
                COMPLEX_TEST_TEXT_AS_UTF_8, options=QRCodeOptions(boost_error=boost_error, image_format="SVG"), cache_enabled=False
            )
            # Using an invalid image format should fallback to SVG.
            url6 = qr_url_from_data(
                COMPLEX_TEST_TEXT_AS_UTF_8, boost_error=boost_error, image_format="invalid-format-name", cache_enabled=False
            )
            url = url1
            urls = get_urls_without_token_for_comparison(url1, url2, url3, url4, url5, url6)
            self.assertEqual(urls[0], urls[1])
            self.assertEqual(urls[0], urls[2])
            self.assertEqual(urls[0], urls[3])
            self.assertEqual(urls[0], urls[4])
            self.assertEqual(urls[0], urls[5])
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            source_image_data = response.content.decode("utf-8")
            ref_file_name = f"{base_file_name}_{str(boost_error).lower()}"
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(ref_file_name, source_image_data)
            ref_image_data = get_svg_content_from_file_name(ref_file_name)
            self.assertEqual(source_image_data, ref_image_data)

    def test_png_boost_error(self):
        base_file_name = "qrfromdata_boost_error"
        for boost_error in [False, True]:
            print("Testing PNG URL with boost_error: %s" % boost_error)
            url1 = make_qr_code_url(
                COMPLEX_TEST_TEXT_AS_UTF_8,
                QRCodeOptions(boost_error=boost_error, image_format="png"),
                force_text=False,
                cache_enabled=False,
            )
            url2 = make_qr_code_url(
                COMPLEX_TEST_TEXT_AS_UTF_8,
                QRCodeOptions(boost_error=boost_error, image_format="PNG"),
                force_text=False,
                cache_enabled=False,
            )
            url3 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, boost_error=boost_error, image_format="png", cache_enabled=False)
            url4 = qr_url_from_data(COMPLEX_TEST_TEXT_AS_UTF_8, boost_error=boost_error, image_format="PNG", cache_enabled=False)
            url5 = qr_url_from_data(
                COMPLEX_TEST_TEXT_AS_UTF_8, options=QRCodeOptions(boost_error=boost_error, image_format="PNG"), cache_enabled=False
            )
            url = url1
            urls = get_urls_without_token_for_comparison(url1, url2, url3, url4, url5)
            self.assertEqual(urls[0], urls[1])
            self.assertEqual(urls[0], urls[2])
            self.assertEqual(urls[0], urls[3])
            self.assertEqual(urls[0], urls[4])
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            source_image_data = response.content
            ref_file_name = f"{base_file_name}_{str(boost_error).lower()}"
            if REFRESH_REFERENCE_IMAGES:
                write_png_content_to_file(ref_file_name, source_image_data)
            ref_image_data = get_png_content_from_file_name(ref_file_name)
            self.assertEqual(source_image_data, ref_image_data)

    def test_svg_encoding(self):
        base_file_name = "qrfromdata_encoding"
        for encoding in ["utf-8", "iso-8859-1"]:
            print("Testing SVG URL with encoding: %s" % encoding)
            url1 = make_qr_code_url(
                COMPLEX_TEST_TEXT.encode(encoding), QRCodeOptions(encoding=encoding), force_text=False, cache_enabled=False
            )
            url2 = qr_url_from_data(COMPLEX_TEST_TEXT.encode(encoding), encoding=encoding, cache_enabled=False)
            url3 = qr_url_from_data(COMPLEX_TEST_TEXT.encode(encoding), encoding=encoding, image_format="svg", cache_enabled=False)
            url4 = qr_url_from_data(COMPLEX_TEST_TEXT.encode(encoding), encoding=encoding, image_format="SVG", cache_enabled=False)
            url5 = qr_url_from_data(
                COMPLEX_TEST_TEXT.encode(encoding), options=QRCodeOptions(encoding=encoding, image_format="SVG"), cache_enabled=False
            )
            # Using an invalid image format should fallback to SVG.
            url6 = qr_url_from_data(
                COMPLEX_TEST_TEXT.encode(encoding), encoding=encoding, image_format="invalid-format-name", cache_enabled=False
            )
            url = url1
            urls = get_urls_without_token_for_comparison(url1, url2, url3, url4, url5, url6)
            self.assertEqual(urls[0], urls[1])
            self.assertEqual(urls[0], urls[2])
            self.assertEqual(urls[0], urls[3])
            self.assertEqual(urls[0], urls[4])
            self.assertEqual(urls[0], urls[5])
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            source_image_data = response.content.decode("utf-8")
            ref_file_name = f"{base_file_name}_{str(encoding).lower()}"
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(ref_file_name, source_image_data)
            ref_image_data = get_svg_content_from_file_name(ref_file_name)
            self.assertEqual(source_image_data, ref_image_data)

    def test_png_encoding(self):
        base_file_name = "qrfromdata_encoding"
        for encoding in ["utf-8", "iso-8859-1"]:
            print("Testing PNG URL with encoding: %s" % encoding)
            url1 = make_qr_code_url(
                COMPLEX_TEST_TEXT.encode(encoding),
                QRCodeOptions(encoding=encoding, image_format="png"),
                force_text=False,
                cache_enabled=False,
            )
            url2 = make_qr_code_url(
                COMPLEX_TEST_TEXT.encode(encoding),
                QRCodeOptions(encoding=encoding, image_format="PNG"),
                force_text=False,
                cache_enabled=False,
            )
            url3 = qr_url_from_data(COMPLEX_TEST_TEXT.encode(encoding), encoding=encoding, image_format="png", cache_enabled=False)
            url4 = qr_url_from_data(COMPLEX_TEST_TEXT.encode(encoding), encoding=encoding, image_format="PNG", cache_enabled=False)
            url5 = qr_url_from_data(
                COMPLEX_TEST_TEXT.encode(encoding), options=QRCodeOptions(encoding=encoding, image_format="PNG"), cache_enabled=False
            )
            url = url1
            urls = get_urls_without_token_for_comparison(url1, url2, url3, url4, url5)
            self.assertEqual(urls[0], urls[1])
            self.assertEqual(urls[0], urls[2])
            self.assertEqual(urls[0], urls[3])
            self.assertEqual(urls[0], urls[4])
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            source_image_data = response.content
            ref_file_name = f"{base_file_name}_{str(encoding).lower()}"
            if REFRESH_REFERENCE_IMAGES:
                write_png_content_to_file(ref_file_name, source_image_data)
            ref_image_data = get_png_content_from_file_name(ref_file_name)
            self.assertEqual(source_image_data, ref_image_data)


class TestQRFromDataSvgResult(SimpleTestCase):
    """
    Ensures that produced QR codes in SVG format coincide with verified references.

    The tests cover direct call to tag function, rendering of tag, and direct call to qr_code API.
    """

    def test_size(self):
        base_ref_file_name = "qrfromdata_size"
        sizes = ["t", "T", "s", "S", None, -1, 0, "m", "M", "l", "L", "h", "H", "6", 6, "8", 8, "10", 10]
        size_names = ["tiny"] * 2 + ["small"] * 2 + ["medium"] * 5 + ["large"] * 2 + ["huge"] * 2 + ["6"] * 2 + ["8"] * 2 + ["10"] * 2
        for i in range(len(sizes)):
            size = sizes[i]
            print("Testing SVG with size %s" % size)
            size_name = size_names[i]
            qr1 = make_embedded_qr_code(TEST_TEXT_AS_UTF_8, QRCodeOptions(size=size), force_text=False)
            qr2 = qr_from_data(TEST_TEXT_AS_UTF_8, size=size)
            qr3 = qr_from_data(TEST_TEXT_AS_UTF_8, size=size, image_format="svg")
            qr4 = qr_from_data(TEST_TEXT_AS_UTF_8, options=QRCodeOptions(size=size, image_format="svg"))
            qr5 = qr_from_data(TEST_TEXT_AS_UTF_8, size=size, image_format="invalid-format-name")
            result_file_name = f"{base_ref_file_name}_{size_name}"
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(result_file_name, qr1)
            result = get_svg_content_from_file_name(result_file_name)
            self.assertEqual(qr1, qr2)
            self.assertEqual(qr1, qr3)
            self.assertEqual(qr1, qr4)
            self.assertEqual(qr1, qr5)
            self.assertEqual(qr1, result)

    def test_version(self):
        base_file_name = "qrfromdata_version"
        versions = [None, -1, 0, 41, "-1", "0", "41", "blabla", 1, "1", 2, "2", 4, "4"]
        version_names = ["default"] * 10 + ["2", "2", "4", "4"]
        for i in range(len(versions)):
            version = versions[i]
            print("Testing SVG with version %s" % version)
            version_name = version_names[i]
            qr1 = make_embedded_qr_code(TEST_TEXT_AS_UTF_8, QRCodeOptions(version=version), force_text=False)
            qr2 = qr_from_data(TEST_TEXT_AS_UTF_8, version=version)
            qr3 = qr_from_data(TEST_TEXT_AS_UTF_8, version=version, image_format="svg")
            qr4 = qr_from_data(TEST_TEXT_AS_UTF_8, version=version, image_format="SVG")
            qr5 = qr_from_data(TEST_TEXT_AS_UTF_8, options=QRCodeOptions(version=version, image_format="SVG"))
            qr6 = qr_from_data(TEST_TEXT_AS_UTF_8, version=version, image_format="invalid-format-name")
            result_file_name = f"{base_file_name}_{version_name}"
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(result_file_name, qr1)
            result = get_svg_content_from_file_name(result_file_name)
            self.assertEqual(qr1, qr2)
            self.assertEqual(qr1, qr3)
            self.assertEqual(qr1, qr4)
            self.assertEqual(qr1, qr5)
            self.assertEqual(qr1, qr6)
            self.assertEqual(qr1, result)

    def test_error_correction(self):
        file_base_name = "qrfromdata_error_correction"
        tests_data = []
        for correction_level in ERROR_CORRECTION_DICT.keys():
            ref_file_name = f"{file_base_name}_{correction_level}"
            tests_data.append(
                dict(
                    source=f'{{% qr_from_data data image_format="svg" error_correction="{correction_level}" %}}',
                    ref_file_name=ref_file_name.lower(),
                )
            )

        for test_data in tests_data:
            print("Testing template: %s" % test_data["source"])
            html_source = mark_safe("{% load qr_code %}" + test_data["source"])
            template = Template(html_source)
            context = Context(dict(data=COMPLEX_TEST_TEXT_AS_UTF_8))
            source_image_data = template.render(context)
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_svg_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(minimal_svg(source_image_data), minimal_svg(ref_image_data))

    def test_eci(self):
        file_base_name = "qrfromdata_eci"
        tests_data = []
        for eci in [False, True]:
            ref_file_name = f"{file_base_name}_{eci}"
            tests_data.append(dict(source=f'{{% qr_from_data data image_format="svg" eci={eci} %}}', ref_file_name=ref_file_name.lower()))

        for test_data in tests_data:
            print("Testing template: %s" % test_data["source"])
            html_source = mark_safe("{% load qr_code %}" + test_data["source"])
            template = Template(html_source)
            context = Context(dict(data=COMPLEX_TEST_TEXT_AS_UTF_8))
            source_image_data = template.render(context)
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_svg_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(minimal_svg(source_image_data), minimal_svg(ref_image_data))

    def test_micro(self):
        file_base_name = "qrfromdata_micro"
        tests_data = []
        for micro in [False, True]:
            ref_file_name = f"{file_base_name}_{micro}"
            tests_data.append(
                dict(
                    source=f'{{% qr_from_data data image_format="svg" micro={micro} encoding="iso-8859-1" %}}',
                    ref_file_name=ref_file_name.lower(),
                )
            )
            tests_data.append(
                dict(
                    source=f'{{% qr_from_data data image_format="svg" micro={micro} encoding=None %}}', ref_file_name=ref_file_name.lower()
                )
            )
            tests_data.append(
                dict(
                    source=f'{{% qr_from_data data image_format="svg" micro={micro} encoding="utf-8" %}}',
                    ref_file_name=ref_file_name.lower(),
                )
            )

        for test_data in tests_data:
            print("Testing template: %s" % test_data["source"])
            html_source = mark_safe("{% load qr_code %}" + test_data["source"])
            template = Template(html_source)
            context = Context(dict(data=COMPLEX_TEST_TEXT_AS_ISO_8859_1))
            source_image_data = template.render(context)
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_svg_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(minimal_svg(source_image_data), minimal_svg(ref_image_data))

    def test_boost_error(self):
        file_base_name = "qrfromdata_boost_error"
        tests_data = []
        for boost_error in [False, True]:
            ref_file_name = f"{file_base_name}_{boost_error}"
            tests_data.append(
                dict(source=f'{{% qr_from_data data image_format="svg" boost_error={boost_error} %}}', ref_file_name=ref_file_name.lower())
            )

        for test_data in tests_data:
            print("Testing template: %s" % test_data["source"])
            html_source = mark_safe("{% load qr_code %}" + test_data["source"])
            template = Template(html_source)
            context = Context(dict(data=COMPLEX_TEST_TEXT_AS_UTF_8))
            source_image_data = template.render(context)
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_svg_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(minimal_svg(source_image_data), minimal_svg(ref_image_data))

    def test_encoding(self):
        file_base_name = "qrfromdata_encoding"
        tests_data = []
        for encoding in ["utf-8", "iso-8859-1"]:
            ref_file_name = f"{file_base_name}_{encoding}"
            tests_data.append(
                dict(
                    encoding=encoding,
                    source=f'{{% qr_from_data data image_format="svg" encoding="{encoding}" %}}',
                    ref_file_name=ref_file_name.lower(),
                )
            )

        for test_data in tests_data:
            print("Testing template: %s" % test_data["source"])
            html_source = mark_safe("{% load qr_code %}" + test_data["source"])
            template = Template(html_source)
            data = COMPLEX_TEST_TEXT.encode(encoding=test_data["encoding"])
            context = Context(dict(data=data))
            source_image_data = template.render(context)
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_svg_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(minimal_svg(source_image_data), minimal_svg(ref_image_data))

    def test_mode(self):
        file_base_name = "qrfromdata_mode"
        tests_data = []
        data_for_mode = [
            TEST_TEXT_AS_UTF_8,
            COMPLEX_TEST_TEXT_AS_UTF_8,
            COMPLEX_TEST_TEXT_AS_ISO_8859_1,
            123,
            "123",
            "ABCD1234",
            "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ $%*+-./:",
            "義務教育諸学校教科用図書検定基準".encode("cp932"),
            "ウェブサイトにおける文字コードの割合、UTF-8が90％超え。Shift_JISやEUC-JPは？".encode("shift-jis"),
        ]
        encodings = ["utf-8", "utf-8", "iso-8859-1", None, None, None, None, None, "shift-jis"]
        for mode_index in range(len(data_for_mode)):
            ref_file_name = f"{file_base_name}_{mode_index}"
            tests_data.append(dict(source=f'{{% qr_from_data data image_format="svg" %}}', ref_file_name=ref_file_name.lower()))

        for i, test_data in enumerate(tests_data):
            print("Testing template: %s" % test_data["source"])
            html_source = mark_safe("{% load qr_code %}" + test_data["source"])
            template = Template(html_source)
            context = Context(dict(data=data_for_mode[i], encoding=encodings[i]))
            source_image_data = template.render(context)
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_svg_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(minimal_svg(source_image_data), minimal_svg(ref_image_data))


class TestQRFromDataPngResult(SimpleTestCase):
    """
    Ensures that produced QR codes in PNG format coincide with verified references.

    The tests cover direct call to tag function, rendering of tag, and direct call to qr_code API.
    """

    def test_size(self):
        base_ref_file_name = "qrfromdata_size"
        sizes = ["t", "T", "s", "S", None, -1, 0, "m", "M", "l", "L", "h", "H", "6", 6, "8", 8, "10", 10]
        size_names = ["tiny"] * 2 + ["small"] * 2 + ["medium"] * 5 + ["large"] * 2 + ["huge"] * 2 + ["6"] * 2 + ["8"] * 2 + ["10"] * 2
        for i in range(len(sizes)):
            size = sizes[i]
            print("Testing PNG with size %s" % size)
            size_name = size_names[i]
            result_file_name = f"{base_ref_file_name}_{size_name}"
            qr1 = make_embedded_qr_code(TEST_TEXT_AS_UTF_8, QRCodeOptions(size=size, image_format="png"), force_text=False)
            qr2 = qr_from_data(TEST_TEXT_AS_UTF_8, size=size, image_format="png")
            qr3 = qr_from_data(TEST_TEXT_AS_UTF_8, options=QRCodeOptions(size=size, image_format="png"), force_text=False)
            if REFRESH_REFERENCE_IMAGES:
                match = IMAGE_TAG_BASE64_DATA_RE.search(qr1)
                source_image_data = match.group("data")
                write_png_content_to_file(result_file_name, base64.b64decode(source_image_data))
            result = base64.b64encode(get_png_content_from_file_name(result_file_name)).decode("utf-8")
            self.assertEqual(qr1, qr2)
            self.assertEqual(qr1, qr3)
            self.assertEqual(qr1, BASE64_PNG_IMAGE_TEMPLATE % result)

    def test_version(self):
        base_file_name = "qrfromdata_version"
        versions = [None, -1, 0, 41, "-1", "0", "41", "blabla", 1, "1", 2, "2", 4, "4"]
        version_names = ["default"] * 10 + ["2", "2", "4", "4"]
        for i in range(len(versions)):
            version = versions[i]
            print("Testing PNG with version %s" % version)
            version_name = version_names[i]
            qr1 = make_embedded_qr_code(TEST_TEXT_AS_UTF_8, QRCodeOptions(version=version, image_format="png"), force_text=False)
            qr2 = qr_from_data(TEST_TEXT_AS_UTF_8, version=version, image_format="png")
            qr3 = qr_from_data(TEST_TEXT_AS_UTF_8, version=version, image_format="PNG")
            qr4 = qr_from_data(TEST_TEXT_AS_UTF_8, options=QRCodeOptions(version=version, image_format="PNG"))
            result_file_name = f"{base_file_name}_{version_name}"
            if REFRESH_REFERENCE_IMAGES:
                match = IMAGE_TAG_BASE64_DATA_RE.search(qr1)
                source_image_data = match.group("data")
                write_png_content_to_file(result_file_name, base64.b64decode(source_image_data))
            result = base64.b64encode(get_png_content_from_file_name(result_file_name)).decode("utf-8")
            self.assertEqual(qr1, qr2)
            self.assertEqual(qr1, qr3)
            self.assertEqual(qr1, qr4)
            self.assertEqual(qr1, BASE64_PNG_IMAGE_TEMPLATE % result)

    def test_error_correction(self):
        file_base_name = "qrfromdata_error_correction"
        tests_data = []
        for correction_level in ERROR_CORRECTION_DICT.keys():
            ref_file_name = f"{file_base_name}_{correction_level}"
            tests_data.append(
                dict(
                    source=f'{{% qr_from_data data image_format="png" error_correction="{correction_level}" %}}',
                    ref_file_name=ref_file_name.lower(),
                )
            )

        for test_data in tests_data:
            print("Testing template: %s" % test_data["source"])
            html_source = mark_safe("{% load qr_code %}" + test_data["source"])
            template = Template(html_source)
            context = Context(dict(data=COMPLEX_TEST_TEXT_AS_UTF_8))
            source_image = template.render(context).strip()
            source_image_data = source_image[32 : -len('" alt="%s"' % escape(COMPLEX_TEST_TEXT))]
            source_image_data = base64.b64decode(source_image_data)
            if REFRESH_REFERENCE_IMAGES:
                write_png_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_png_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(source_image_data, ref_image_data)

    def test_eci(self):
        file_base_name = "qrfromdata_eci"
        tests_data = []
        for eci in [False, True]:
            ref_file_name = f"{file_base_name}_{eci}"
            tests_data.append(dict(source=f'{{% qr_from_data data image_format="png" eci={eci} %}}', ref_file_name=ref_file_name.lower()))
            # tests_data.append(dict(source=f'{{% qr_from_data "{COMPLEX_TEST_TEXT_AS_UTF_8}" image_format="png" eci="{eci}" %}}', ref_file_name=ref_file_name.lower()))

        for test_data in tests_data:
            print("Testing template: %s" % test_data["source"])
            html_source = mark_safe("{% load qr_code %}" + test_data["source"])
            template = Template(html_source)
            context = Context(dict(data=COMPLEX_TEST_TEXT_AS_UTF_8))
            source_image = template.render(context).strip()
            source_image_data = source_image[32 : -len('" alt="%s"' % escape(COMPLEX_TEST_TEXT))]
            source_image_data = base64.b64decode(source_image_data)
            if REFRESH_REFERENCE_IMAGES:
                write_png_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_png_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(source_image_data, ref_image_data)

    def test_micro(self):
        file_base_name = "qrfromdata_micro"
        tests_data = []
        for micro in [False, True]:
            ref_file_name = f"{file_base_name}_{micro}"
            tests_data.append(
                dict(source=f'{{% qr_from_data data image_format="png"  micro={micro}%}}', ref_file_name=ref_file_name.lower())
            )

        for test_data in tests_data:
            print("Testing template: %s" % test_data["source"])
            html_source = mark_safe("{% load qr_code %}" + test_data["source"])
            template = Template(html_source)
            context = Context(dict(data=COMPLEX_TEST_TEXT_AS_ISO_8859_1))
            source_image = template.render(context).strip()
            source_image_data = source_image[32 : -len('" alt="%s"' % escape(COMPLEX_TEST_TEXT))]
            source_image_data = base64.b64decode(source_image_data)
            if REFRESH_REFERENCE_IMAGES:
                write_png_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_png_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(source_image_data, ref_image_data)

    def test_boost_error(self):
        file_base_name = "qrfromdata_boost_error"
        tests_data = []
        for boost_error in [False, True]:
            ref_file_name = f"{file_base_name}_{boost_error}"
            tests_data.append(
                dict(source=f'{{% qr_from_data data image_format="png" boost_error={boost_error}%}}', ref_file_name=ref_file_name.lower())
            )

        for test_data in tests_data:
            print("Testing template: %s" % test_data["source"])
            html_source = mark_safe("{% load qr_code %}" + test_data["source"])
            template = Template(html_source)
            context = Context(dict(data=COMPLEX_TEST_TEXT_AS_UTF_8))
            source_image = template.render(context).strip()
            source_image_data = source_image[32 : -len('" alt="%s"' % escape(COMPLEX_TEST_TEXT))]
            source_image_data = base64.b64decode(source_image_data)
            if REFRESH_REFERENCE_IMAGES:
                write_png_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_png_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(source_image_data, ref_image_data)

    def test_encoding(self):
        file_base_name = "qrfromdata_encoding"
        tests_data = []
        for encoding in ["utf-8", "iso-8859-1"]:
            ref_file_name = f"{file_base_name}_{encoding}"
            tests_data.append(
                dict(
                    encoding=encoding,
                    source=f'{{% qr_from_data data image_format="png" encoding="{encoding}" %}}',
                    ref_file_name=ref_file_name.lower(),
                )
            )

        for test_data in tests_data:
            print("Testing template: %s" % test_data["source"])
            html_source = mark_safe("{% load qr_code %}" + test_data["source"])
            template = Template(html_source)
            data = COMPLEX_TEST_TEXT.encode(encoding=test_data["encoding"])
            context = Context(dict(data=data))
            source_image = template.render(context).strip()
            source_image_data = source_image[32 : -len('" alt="%s"' % escape(COMPLEX_TEST_TEXT))]
            source_image_data = base64.b64decode(source_image_data)
            if REFRESH_REFERENCE_IMAGES:
                write_png_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_png_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(source_image_data, ref_image_data)

    def test_mode(self):
        file_base_name = "qrfromdata_mode"
        tests_data = []
        data_for_mode = [
            TEST_TEXT_AS_UTF_8,
            COMPLEX_TEST_TEXT_AS_UTF_8,
            COMPLEX_TEST_TEXT_AS_ISO_8859_1,
            123,
            "123",
            "ABCD1234",
            "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ $%*+-./:",
            "義務教育諸学校教科用図書検定基準".encode("cp932"),
            "ウェブサイトにおける文字コードの割合、UTF-8が90％超え。Shift_JISやEUC-JPは？".encode("shift-jis"),
        ]
        encodings = ["utf-8", "utf-8", "iso-8859-1", None, None, None, None, None, "shift-jis"]
        for mode_index in range(len(data_for_mode)):
            ref_file_name = f"{file_base_name}_{mode_index}"
            tests_data.append(
                dict(source=f'{{% qr_from_data data image_format="png" encoding=encoding %}}', ref_file_name=ref_file_name.lower())
            )

        for i, test_data in enumerate(tests_data):
            print("Testing template: %s" % test_data["source"])
            html_source = mark_safe("{% load qr_code %}" + test_data["source"])
            template = Template(html_source)
            context = Context(dict(data=data_for_mode[i], encoding=encodings[i]))
            source_image = template.render(context).strip()
            source_image_data = source_image[32 : source_image.index('" alt="')]
            source_image_data = base64.b64decode(source_image_data)
            if REFRESH_REFERENCE_IMAGES:
                write_png_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_png_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(source_image_data, ref_image_data)



"""Tools for generating QR codes. This module depends on the Segno library."""
import io
from typing import Mapping, Any

from django.utils.html import escape
from django.utils.safestring import mark_safe
import segno
from pydantic import validate_arguments

from qr_code.qrcode import PYDANTIC_CONFIG
from qr_code.qrcode.constants import DEFAULT_CACHE_ENABLED, DEFAULT_URL_SIGNATURE_ENABLED
from qr_code.qrcode.serve import make_qr_code_url
from qr_code.qrcode.utils import QRCodeOptions


@validate_arguments(config=PYDANTIC_CONFIG)
def make_qr(data: Any, qr_code_options: QRCodeOptions, force_text: bool = True):
    """Creates a QR code that encodes the given `data` with the given `qr_code_options`.

    :param str data: The data to encode
    :param qr_code_options: Options to create and serialize the QR code.
    :param bool force_text: Tells whether we want to force the `data` to be considered as text string and encoded in byte mode.
    :rtype: segno.QRCode
    """
    # WARNING: For compatibility reasons, we still allow to pass __proxy__ class (lazy string). Moreover, it would be OK to pass anything that has __str__
    # attribute (e. g. class instance that handles phone numbers).
    if force_text:
        return segno.make(str(data), **qr_code_options.kw_make(), mode="byte")
    return segno.make(data, **qr_code_options.kw_make())


@validate_arguments(config=PYDANTIC_CONFIG)
def make_qr_code_image(data: Any, qr_code_options: QRCodeOptions, force_text: bool = True) -> bytes:
    """
    Creates a bytes object representing a QR code image for the provided `data`.

    :param str data: The data to encode
    :param qr_code_options: Options to create and serialize the QR code.
    :param bool force_text: Tells whether we want to force the `data` to be considered as text string and encoded in byte mode.
    :rtype: bytes
    """
    qr = make_qr(data, qr_code_options, force_text=force_text)
    out = io.BytesIO()
    qr.save(out, **qr_code_options.kw_save())
    return out.getvalue()


@validate_arguments(config=PYDANTIC_CONFIG)
def make_embedded_qr_code(data: Any, qr_code_options: QRCodeOptions, force_text: bool = True) -> str:
    """
    Generates a <svg> or <img> tag representing the QR code for the given `data`.
    This tag can be embedded into an HTML document.
    """
    qr = make_qr(data, qr_code_options, force_text=force_text)
    kw = qr_code_options.kw_save()
    # Pop the image format from the keywords since qr.png_data_uri / qr.svg_inline
    # set it automatically
    kw.pop("kind")
    if qr_code_options.image_format == "png":
        if isinstance(data, bytes):
            alt = ""
            encodings = ["utf-8", "iso-8859-1", "shift-jis"]
            if qr_code_options.encoding:
                ei = encodings.index(qr_code_options.encoding)
                if ei > 0:
                    encodings[ei] = encodings[0]
                    encodings[0] = qr_code_options.encoding
            for e in encodings:
                try:
                    alt = data.decode(e)
                    break
                except UnicodeDecodeError:
                    pass
        elif not isinstance(data, str):
            alt = str(data)
        else:
            alt = data
        return mark_safe(f'<img src="{qr.png_data_uri(**kw)}" alt="{escape(alt)}">')
    return mark_safe(qr.svg_inline(**kw))


def make_qr_code_with_args(data: Any, qr_code_args: dict, force_text: bool = True) -> str:
    options = _options_from_args(qr_code_args)
    return make_embedded_qr_code(data, options, force_text=force_text)


def make_qr_code_url_with_args(data: Any, qr_code_args: dict, force_text: bool = True) -> str:
    cache_enabled = qr_code_args.pop("cache_enabled", DEFAULT_CACHE_ENABLED)
    if not isinstance(cache_enabled, bool):
        cache_enabled = not cache_enabled == "False"
    url_signature_enabled = qr_code_args.pop("url_signature_enabled", DEFAULT_URL_SIGNATURE_ENABLED)
    if not isinstance(url_signature_enabled, bool):
        url_signature_enabled = not url_signature_enabled == "False"
    options = _options_from_args(qr_code_args)
    return make_qr_code_url(data, options, force_text=force_text, cache_enabled=cache_enabled, url_signature_enabled=url_signature_enabled)


def _options_from_args(args: Mapping) -> QRCodeOptions:
    """Returns a QRCodeOptions instance from the provided arguments."""
    options = args.get("options")
    if options:
        if not isinstance(options, QRCodeOptions):
            raise TypeError("The options argument must be of type QRCodeOptions.")
    else:
        # Convert the string "None" into None
        kw = {k: v if v != "None" else None for k, v in args.items()}
        options = QRCodeOptions(**kw)
    return options



"""Tests for qr_code application."""
import base64
import datetime

from dataclasses import asdict
from datetime import date

import pytz
from django.template import Template, Context
from django.test import SimpleTestCase, override_settings
from django.utils.safestring import mark_safe

from qr_code.qrcode.utils import (
    ContactDetail,
    VCard,
    MeCard,
    WifiConfig,
    Coordinates,
    EpcData,
    VEvent,
    EventClass,
    EventTransparency,
    EventStatus,
)
from qr_code.tests import REFRESH_REFERENCE_IMAGES, IMAGE_TAG_BASE64_DATA_RE
from qr_code.tests.utils import (
    write_svg_content_to_file,
    get_svg_content_from_file_name,
    minimal_svg,
    get_png_content_from_file_name,
    write_png_content_to_file,
)

US_EASTERN_TZ = pytz.timezone("US/Eastern")
EUROPE_ZURICH_TZ = pytz.timezone("Europe/Zurich")
TEST_EVENT1 = VEvent(
    uid="django-qr-code-test-id-1",
    summary="Vacations",
    start=US_EASTERN_TZ.localize(datetime.datetime(2022, 7, 6, hour=8, minute=30)),
    end=US_EASTERN_TZ.localize(datetime.datetime(2022, 7, 17, hour=12)),
    location="New-York",
    categories=["holidays"],
    event_class=EventClass.PUBLIC,
    transparency=EventTransparency.TRANSPARENT,
    dtstamp=datetime.datetime(2022, 6, 25, hour=17, minute=30, tzinfo=pytz.timezone("UTC")),
)
TEST_EVENT2 = VEvent(
    uid="django-qr-code-test-id-2",
    summary="Café avec Marcel!",
    start=EUROPE_ZURICH_TZ.localize(datetime.datetime(2022, 6, 27, hour=8, minute=15)),
    end=EUROPE_ZURICH_TZ.localize(datetime.datetime(2022, 6, 27, hour=9)),
    categories=["PERSO,FRIENDS"],
    event_class=EventClass.PRIVATE,
    dtstamp=datetime.datetime(2022, 6, 25, hour=17, minute=30, tzinfo=pytz.timezone("UTC")),
)
TEST_EVENT3 = VEvent(
    uid="django-qr-code-test-id-3",
    summary="Vacations",
    start=US_EASTERN_TZ.localize(datetime.datetime(2022, 8, 6, hour=8, minute=30)),
    end=US_EASTERN_TZ.localize(datetime.datetime(2022, 8, 17, hour=12)),
    location="New-York",
    categories=["holidays", "PERSONAL", "FAMILY_STUFF", "FUN AT HOME", "SOME_VERY_LONG_CATEGORY_NAME_THAT_REQUIRES_LINE_FOLDING"],
    status=EventStatus.TENTATIVE,
    organizer="foo@bar.com",
    url="https://bar.com",
    description="""Meeting to provide technical review for "Phoenix" design.
Happy Face Conference Room.
Phoenix design team MUST attend this meeting.
RSVP to team leader.""",
    dtstamp=datetime.datetime(2022, 6, 25, hour=17, minute=30, tzinfo=pytz.timezone("UTC")),
)
TEST_EVENT4 = VEvent(
    uid="django-qr-code-test-id-4",
    summary="Concert",
    start=datetime.datetime(2022, 8, 6, hour=19, minute=30),
    end=datetime.datetime(2022, 8, 6, hour=23),
    location="Delémont, Salle St-Georges",
    geo=(47.3661640157025, 7.346173125268345),
    status=EventStatus.CONFIRMED,
    description="""Meeting to provide technical review for "Phoenix" design. Happy Face Conference Room. Phoenix design team MUST attend this meeting.

RSVP to team leader.

Add some diacritics for fun: éàüî""",
    dtstamp=datetime.datetime(2022, 6, 25, hour=17, minute=30, tzinfo=pytz.timezone("UTC")),
)
TEST_CONTACT_DETAIL = dict(
    first_name="Jérémy Sébastien Ninõ",
    last_name="Érard",
    first_name_reading="jAAn",
    last_name_reading="dOH",
    tel="+41769998877",
    email="j.doe@company.com",
    url="http://www.company.com",
    birthday=date(year=1985, month=10, day=2),
    address="Cras des Fourches 987, 2800 Delémont, Jura, Switzerland",
    memo="Development Manager",
    org="Company Ltd",
)
TEST_MECARD_CONTACT = MeCard(
    name="Ninõ,Jérémy Sébastien",
    phone="+41769998877",
    email="j.doe@company.com",
    url="http://www.company.com",
    birthday=date(year=1985, month=10, day=2),
    memo="Development Manager",
    org="Company Ltd",
)

TEST_VCARD_CONTACT = VCard(
    name="Ninõ;Jérémy Sébastien",
    phone="+41769998877",
    email="j.doe@company.com",
    url="http://www.company.com",
    birthday=date(year=1985, month=10, day=2),
    street="Cras des Fourches 987",
    city="Delémont",
    zipcode=2800,
    region="Jura",
    country="Switzerland",
    memo="Development Manager",
    org="Company Ltd",
)
TEST_VCARD_CONTACT2 = VCard(
    name="Ninõ;Jérémy Sébastien",
    phone="+41769998877",
    email="j.doe@company.com",
    url="http://www.company.com",
    birthday=date(year=1985, month=10, day=2),
    street="Cras des Fourches 987",
    city="Delémont",
    zipcode=2800,
    region="Jura",
    country="Switzerland",
    memo="Development Manager",
    org="Company Ltd",
    photo_uri="http://www.company.com/profile/12",
    cellphone=["+41769998877", "+41769998878"],
    homephone="+41321112233",
    workphone="+41329992233",
)
TEST_WIFI_CONFIG = dict(ssid="my-wifi", authentication=WifiConfig.AUTHENTICATION.WPA, password="wifi-password")
TEST_EPC_QR_1 = dict(
    name="Wikimedia Foerdergesellschaft", iban="DE33100205000001194700", amount=20, text="To Wikipedia, From Gérard Boéchat"
)
TEST_EPC_QR_2 = dict(name="Wikimedia Foerdergesellschaft", iban="DE33100205000001194700", amount=50.0, reference="12983020")


class TestContactDetail(SimpleTestCase):
    def test_make_qr_code_text_with_contact_detail(self):
        data = dict(**TEST_CONTACT_DETAIL)
        c1 = ContactDetail(**data)
        data["nickname"] = "buddy"
        c2 = ContactDetail(**data)
        data["last_name"] = "O'Hara;,:"
        data["tel_av"] = "n/a"
        c3 = ContactDetail(**data)
        del data["last_name"]
        c4 = ContactDetail(**data)
        self.assertEqual(
            c1.make_qr_code_data(),
            r"MECARD:N:Érard,Jérémy Sébastien Ninõ;SOUND:dOH,jAAn;TEL:+41769998877;EMAIL:j.doe@company.com;NOTE:Development Manager;BDAY:19851002;ADR:Cras des Fourches 987, 2800 Delémont, Jura, Switzerland;URL:http\://www.company.com;ORG:Company Ltd;;",
        )
        self.assertEqual(
            c2.make_qr_code_data(),
            r"MECARD:N:Érard,Jérémy Sébastien Ninõ;SOUND:dOH,jAAn;TEL:+41769998877;EMAIL:j.doe@company.com;NOTE:Development Manager;BDAY:19851002;ADR:Cras des Fourches 987, 2800 Delémont, Jura, Switzerland;URL:http\://www.company.com;NICKNAME:buddy;ORG:Company Ltd;;",
        )
        self.assertEqual(
            c3.make_qr_code_data(),
            r"MECARD:N:O'Hara\;\,\:,Jérémy Sébastien Ninõ;SOUND:dOH,jAAn;TEL:+41769998877;TEL-AV:n/a;EMAIL:j.doe@company.com;NOTE:Development Manager;BDAY:19851002;ADR:Cras des Fourches 987, 2800 Delémont, Jura, Switzerland;URL:http\://www.company.com;NICKNAME:buddy;ORG:Company Ltd;;",
        )
        self.assertEqual(
            c4.make_qr_code_data(),
            r"MECARD:N:Jérémy Sébastien Ninõ;SOUND:dOH,jAAn;TEL:+41769998877;TEL-AV:n/a;EMAIL:j.doe@company.com;NOTE:Development Manager;BDAY:19851002;ADR:Cras des Fourches 987, 2800 Delémont, Jura, Switzerland;URL:http\://www.company.com;NICKNAME:buddy;ORG:Company Ltd;;",
        )

    def test_make_qr_code_text_with_mecard(self):
        data = asdict(TEST_MECARD_CONTACT)
        c1 = MeCard(**data)
        data["nickname"] = "buddy"
        c2 = MeCard(**data)
        self.assertEqual(
            c1.make_qr_code_data(),
            r"MECARD:N:Ninõ,Jérémy Sébastien;TEL:+41769998877;EMAIL:j.doe@company.com;BDAY:19851002;URL:http\://www.company.com;MEMO:Development Manager;;ORG:Company Ltd;",
        )
        self.assertEqual(
            c2.make_qr_code_data(),
            r"MECARD:N:Ninõ,Jérémy Sébastien;TEL:+41769998877;EMAIL:j.doe@company.com;NICKNAME:buddy;BDAY:19851002;URL:http\://www.company.com;MEMO:Development Manager;;ORG:Company Ltd;",
        )

    def test_make_qr_code_text_with_vcard(self):
        data = asdict(TEST_VCARD_CONTACT)
        c1 = VCard(**data)
        data["nickname"] = "buddy"
        c2 = VCard(**data)
        self.assertEqual(
            c1.make_qr_code_data(),
            """BEGIN:VCARD\r
VERSION:3.0\r
N:Ninõ;Jérémy Sébastien\r
FN:Ninõ Jérémy Sébastien\r
ORG:Company Ltd\r
EMAIL:j.doe@company.com\r
TEL:+41769998877\r
URL:http://www.company.com\r
ADR:;;Cras des Fourches 987;Delémont;Jura;2800;Switzerland\r
BDAY:1985-10-02\r
NOTE:Development Manager\r
END:VCARD\r
""",
        )
        self.assertEqual(
            c2.make_qr_code_data(),
            """BEGIN:VCARD\r
VERSION:3.0\r
N:Ninõ;Jérémy Sébastien\r
FN:Ninõ Jérémy Sébastien\r
ORG:Company Ltd\r
EMAIL:j.doe@company.com\r
TEL:+41769998877\r
URL:http://www.company.com\r
NICKNAME:buddy\r
ADR:;;Cras des Fourches 987;Delémont;Jura;2800;Switzerland\r
BDAY:1985-10-02\r
NOTE:Development Manager\r
END:VCARD\r
""",
        )


class TestWifiConfig(SimpleTestCase):
    def test_make_qr_code_text(self):
        wifi1 = WifiConfig(**TEST_WIFI_CONFIG)
        wifi2 = WifiConfig(hidden=True, **TEST_WIFI_CONFIG)
        self.assertEqual(wifi1.make_qr_code_data(), "WIFI:S:my-wifi;T:WPA;P:wifi-password;;")
        self.assertEqual(wifi2.make_qr_code_data(), "WIFI:S:my-wifi;T:WPA;P:wifi-password;H:true;;")


class TestCoordinates(SimpleTestCase):
    def test_coordinates(self):
        c1 = Coordinates(latitude=586000.32, longitude=250954.19)
        c2 = Coordinates(latitude=586000.32, longitude=250954.19, altitude=500)
        self.assertEqual(c1.__str__(), "latitude: 586000.32, longitude: 250954.19")
        self.assertEqual(c2.__str__(), "latitude: 586000.32, longitude: 250954.19, altitude: 500")


class TestQRForApplications(SimpleTestCase):
    @staticmethod
    def _make_test_data(tag_pattern, ref_file_name, tag_args, template_context=dict()):
        tag_content = tag_pattern
        for key, value in tag_args.items():
            if isinstance(value, str):
                tag_content += f' {key}="{value}"'
            else:
                tag_content += f" {key}={value}"
        return dict(source="{% " + tag_content + " %}", ref_file_name=ref_file_name, template_context=template_context)

    @staticmethod
    def _make_tests_data(embedded=True, image_format="svg"):
        contact_detail1 = dict(**TEST_CONTACT_DETAIL)
        contact_detail2 = ContactDetail(**contact_detail1)
        wifi_config1 = dict(**TEST_WIFI_CONFIG)
        wifi_config2 = WifiConfig(**wifi_config1)
        epc_data1 = dict(**TEST_EPC_QR_1)
        epc_data2 = EpcData(**epc_data1)
        epc_data3 = dict(**TEST_EPC_QR_2)
        google_maps_coordinates = Coordinates(latitude=586000.32, longitude=250954.19)
        geolocation_coordinates = Coordinates(latitude=586000.32, longitude=250954.19, altitude=500)
        tag_prefix = "qr_for_" if embedded else "qr_url_for_"
        tag_args = dict(image_format=image_format)
        if image_format == "png":
            tag_args["size"] = "t"
        if not embedded:
            # Deactivate cache for URL.
            tag_args["cache_enabled"] = False
        raw_data = (
            ("email", '"john.doe@domain.com"', None, None),
            ("tel", ' "+41769998877"', None, None),
            ("sms", ' "+41769998877"', None, None),
            ("geolocation", "latitude=586000.32 longitude=250954.19 altitude=500", None, None),
            ("geolocation", "coordinates=coordinates", {"coordinates": geolocation_coordinates}, None),
            ("google_maps", "latitude=586000.32 longitude=250954.19", None, None),
            ("google_maps", "coordinates=coordinates", {"coordinates": google_maps_coordinates}, None),
            ("wifi", "wifi_config", {"wifi_config": wifi_config1}, None),
            ("wifi", "wifi_config", {"wifi_config": wifi_config2}, None),
            ("wifi", "wifi_config=wifi_config", {"wifi_config": wifi_config2}, None),
            ("epc", "epc_data", {"epc_data": epc_data1}, 1),
            ("epc", "epc_data", {"epc_data": epc_data2}, 1),
            ("epc", "epc_data=epc_data", {"epc_data": epc_data2}, 1),
            ("epc", "epc_data", {"epc_data": epc_data3}, 2),
            ("contact", "contact_detail", {"contact_detail": contact_detail1}, None),
            ("contact", "contact_detail", {"contact_detail": contact_detail2}, None),
            ("contact", "contact_detail=contact_detail", {"contact_detail": contact_detail2}, None),
            ("mecard", "mecard=mecard", {"mecard": TEST_MECARD_CONTACT}, None),
            ("mecard", "mecard", {"mecard": TEST_MECARD_CONTACT}, None),
            ("vcard", "vcard=vcard", {"vcard": TEST_VCARD_CONTACT}, 1),
            ("vcard", "vcard", {"vcard": TEST_VCARD_CONTACT}, 1),
            ("vcard", "vcard", {"vcard": TEST_VCARD_CONTACT2}, 2),
            ("event", "event", {"event": TEST_EVENT1}, 1),
            ("event", "event", {"event": TEST_EVENT2}, 2),
            ("event", "event", {"event": TEST_EVENT3}, 3),
            ("event", "event", {"event": TEST_EVENT4}, 4),
            ("youtube", '"J9go2nj6b3M"', None, None),
            ("youtube", "video_id", {"video_id": "J9go2nj6b3M"}, None),
            ("google_play", '"ch.admin.meteoswiss"', None, None),
        )
        tests_data = []
        for tag_base_name, tag_data, template_context, number in raw_data:
            ref_file_name = "qr_for_%s" % tag_base_name
            if number is not None:
                ref_file_name = f"{ref_file_name}_{number}"
            test_data = TestQRForApplications._make_test_data(
                tag_pattern=f"{tag_prefix}{tag_base_name} {tag_data}",
                ref_file_name=ref_file_name,
                tag_args=tag_args,
                template_context=template_context,
            )
            tests_data.append(test_data)
        return tests_data

    @staticmethod
    def _get_rendered_template(template_source, template_context):
        html_source = mark_safe("{% load qr_code %}" + template_source)
        template = Template(html_source)
        context = Context()
        if template_context:
            context.update(template_context)
        return template.render(context).strip()

    def test_demo_samples_embedded_in_svg_format(self):
        tests_data = self._make_tests_data(embedded=True)
        for test_data in tests_data:
            print("Testing template: %s" % test_data["source"])
            source_image_data = TestQRForApplications._get_rendered_template(test_data["source"], test_data.get("template_context"))
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_svg_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(source_image_data, ref_image_data)

    def test_demo_samples_embedded_in_png_format(self):
        tests_data = self._make_tests_data(embedded=True, image_format="png")
        for test_data in tests_data:
            print("Testing template: %s" % test_data["source"])
            source_image_data = TestQRForApplications._get_rendered_template(test_data["source"], test_data.get("template_context"))
            match = IMAGE_TAG_BASE64_DATA_RE.search(source_image_data)
            source_image_data = match.group("data")
            source_image_data = base64.b64decode(source_image_data)
            if REFRESH_REFERENCE_IMAGES:
                write_png_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_png_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(source_image_data, ref_image_data)

    def test_demo_sample_urls_in_svg_format(self):
        tests_data = self._make_tests_data(embedded=False)
        for test_data in tests_data:
            source_image_data = self._check_url_for_test_data(test_data).content.decode("utf-8")
            if REFRESH_REFERENCE_IMAGES:
                write_svg_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_svg_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(minimal_svg(source_image_data), minimal_svg(ref_image_data))

    def test_demo_sample_urls_in_png_format(self):
        tests_data = self._make_tests_data(embedded=False, image_format="png")
        for test_data in tests_data:
            source_image_data = self._check_url_for_test_data(test_data).content
            if REFRESH_REFERENCE_IMAGES:
                write_png_content_to_file(test_data["ref_file_name"], source_image_data)
            ref_image_data = get_png_content_from_file_name(test_data["ref_file_name"])
            self.assertEqual(source_image_data, ref_image_data)

    def _check_url_for_test_data(self, test_data):
        print("Testing template: %s" % test_data["source"])
        source_image_url = TestQRForApplications._get_rendered_template(test_data["source"], test_data.get("template_context"))
        response = self.client.get(source_image_url)
        self.assertEqual(response.status_code, 200)
        return response



from django.urls import path, include

from qr_code_demo import views

app_name = "qr_code_demo"
urlpatterns = [
    path("", views.index, name="index"),
    path("qr-code/", include("qr_code.urls", namespace="qr_code")),
]



"""
Django settings for qr_code_demo project.

Generated by 'django-admin startproject' using Django 2.2.

For more information on this file, see
https://docs.djangoproject.com/en/2.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.2/ref/settings/
"""

import os
import django

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
from qr_code.qrcode import constants

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "8l4)()f1&tg*dtxh6whlew#k-d5&79npe#j_dg9l0b)m8^g#8u"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []

# Application definition

INSTALLED_APPS = ["django.contrib.auth", "django.contrib.contenttypes", "django.contrib.staticfiles", "qr_code", "qr_code_demo"]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "demo_site.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "demo_site.wsgi.application"

# Database
# https://docs.djangoproject.com/en/3.1/ref/settings/#databases

DATABASES = {}


# Password validation
# https://docs.djangoproject.com/en/3.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]

# Internationalization
# https://docs.djangoproject.com/en/3.1/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.1/howto/static-files/

STATIC_URL = "/static/"
STATIC_ROOT = os.path.join(BASE_DIR, "staticfiles")

# Caches.
CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
    },
    "qr-code": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache", "LOCATION": "qr-code-cache", "TIMEOUT": 3600},
}

# Django QR Code specific options.
QR_CODE_CACHE_ALIAS = "qr-code"
QR_CODE_URL_PROTECTION = {
    constants.TOKEN_LENGTH: 30,  # Optional random token length for URL protection. Defaults to 20.
    constants.SIGNING_KEY: "my-secret-signing-key",  # Optional signing key for URL token. Uses SECRET_KEY if not defined.
    constants.SIGNING_SALT: "my-signing-salt",  # Optional signing salt for URL token.
    constants.ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER: lambda u: True,  # Tells whether a registered user can request the QR code URLs from outside a site that uses this app. It can be a boolean value used for any user, or a callable that takes a user as parameter. Defaults to False (nobody can access the URL without the security token).
}
SERVE_QR_CODE_IMAGE_PATH = "qr-code-image/"



from django.conf.urls import include
from django.urls import path
from django.views.generic import RedirectView

urlpatterns = [
    path("", RedirectView.as_view(url="qr-code-demo/", permanent=True)),
    path("qr-code-demo/", include("qr_code_demo.urls", namespace="qr_code_demo")),
    path("qr-code/", include("qr_code.urls", namespace="qr_code")),
]



import re

BASE64_PNG_IMAGE_TEMPLATE = '<img src="data:image/png;base64,%s" alt="Hello World!">'
IMAGE_TAG_BASE64_DATA_RE = re.compile(r"data:image/png;base64,(?P<data>[\w/+=]+)")
TEST_TEXT = "Hello World!"
COMPLEX_TEST_TEXT = "/%+¼@#=<>àé"

OVERRIDE_CACHES_SETTING = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
    },
    "qr-code": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache", "LOCATION": "qr-code-cache", "TIMEOUT": 3600},
}
SVG_REF_SUFFIX = ".ref.svg"
PNG_REF_SUFFIX = ".ref.png"

# Set this flag to True for writing the new version of each reference image in tests/resources while running the tests.
REFRESH_REFERENCE_IMAGES = False



import base64
import binascii
import functools

from django.conf import settings
from django.core.exceptions import PermissionDenied, SuspiciousOperation
from django.core.signing import BadSignature, Signer
from django.http import HttpResponse
from django.views.decorators.cache import cache_page
from django.views.decorators.http import condition

from qr_code.qrcode import constants
from qr_code.qrcode.maker import make_qr_code_image
from qr_code.qrcode.utils import QRCodeOptions
from qr_code.qrcode.serve import (
    get_url_protection_options,
    get_qr_url_protection_token,
    qr_code_etag,
    qr_code_last_modified,
    allows_external_request_from_user,
)


def cache_qr_code():
    """
    Decorator that caches the requested page if a settings named 'QR_CODE_CACHE_ALIAS' exists and is not empty or None.
    """

    def decorator(view_func):
        @functools.wraps(view_func)
        def _wrapped_view(request, *view_args, **view_kwargs):
            cache_enabled = int(request.GET.get("cache_enabled", 1)) == 1
            if cache_enabled and hasattr(settings, "QR_CODE_CACHE_ALIAS") and settings.QR_CODE_CACHE_ALIAS:
                # We found a cache alias for storing the generate qr code and cache is enabled, use it to cache the
                # page.
                timeout = settings.CACHES[settings.QR_CODE_CACHE_ALIAS]["TIMEOUT"]
                key_prefix = "token={}.user_pk={}".format(
                    request.GET.get("url_signature_enabled") or constants.DEFAULT_URL_SIGNATURE_ENABLED,
                    request.user.pk,
                )
                response = cache_page(timeout, cache=settings.QR_CODE_CACHE_ALIAS, key_prefix=key_prefix)(view_func)(
                    request, *view_args, **view_kwargs
                )
            else:
                # No cache alias for storing the generated qr code, call the view as is.
                response = (view_func)(request, *view_args, **view_kwargs)
            return response

        return _wrapped_view

    return decorator


@condition(etag_func=qr_code_etag, last_modified_func=qr_code_last_modified)
@cache_qr_code()
def serve_qr_code_image(request) -> HttpResponse:
    """Serve an image that represents the requested QR code.

    IMPORTANT NOTE: Each boolean option mentioned below (value True/False) must be passed as `<option name>=1` for `True`, respectively `<option name>=0` for `False`.

    You may pass any argument available for :py:func:`qr_code.qrcode.utils.QRCodeOptions.__init__` to adjust the appearance of the returned QR code. The arguments must be passed as query
    string arguments in the URL. Additionally, the following arguments are available:
        * cache_enabled: boolean – Tells the generator to work around the caching mechanism if it is empty (default: True/undefined).
        * token: str – By default, the application only serves QR code images for authenticated URLs (requests generated from your application and addressed to your application).
            The authentication uses a HMAC to sign the request query arguments. The authentication code is passed as a query argument named `token` which is automatically generated
            by `qr_url_from_text` or `qr_url_from_data`.
    """
    qr_code_options = get_qr_code_option_from_request(request)
    # Handle image access protection (we do not allow external requests for anyone).
    check_image_access_permission(request, qr_code_options)
    force_text = False
    if "bytes" in request.GET:
        try:
            data = base64.b64decode(request.GET.get("bytes", b""))
        except binascii.Error:
            raise SuspiciousOperation("Invalid base64 encoded data.")
    elif "int" in request.GET:
        try:
            data = int(request.GET.get("int", None))  # type: ignore
        except (ValueError, TypeError):
            raise SuspiciousOperation("Invalid integer value.")
    else:
        try:
            data = base64.b64decode(request.GET.get("text", "")).decode("utf-8")  # type: ignore
            force_text = True
        except binascii.Error:
            raise SuspiciousOperation("Invalid base64 encoded text.")
        except UnicodeDecodeError:
            raise SuspiciousOperation("Invalid UTF-8 encoded text.")
    img = make_qr_code_image(data, qr_code_options=qr_code_options, force_text=force_text)
    return HttpResponse(content=img, content_type="image/svg+xml" if qr_code_options.image_format == "svg" else "image/png")


def get_qr_code_option_from_request(request) -> QRCodeOptions:
    request_query = request.GET.dict()
    for key in ("bytes", "text", "int", "token", "cache_enabled"):
        request_query.pop(key, None)
    # Force typing for booleans.
    request_query["micro"] = int(request_query.get("micro", 0)) == 1
    request_query["eci"] = int(request_query.get("eci", 0)) == 1
    request_query["boost_error"] = int(request_query.get("boost_error", 0)) == 1
    return QRCodeOptions(**request_query)


def check_image_access_permission(request, qr_code_options) -> None:
    """Handle image access protection (we do not allow external requests for anyone)."""
    token = request.GET.get("token", "")
    if token:
        check_url_signature_token(qr_code_options, token)
    else:
        if not allows_external_request_from_user(request.user):
            raise PermissionDenied("You are not allowed to access this QR code.")


def check_url_signature_token(qr_code_options, token) -> None:
    url_protection_options = get_url_protection_options()
    signer = Signer(key=url_protection_options[constants.SIGNING_KEY], salt=url_protection_options[constants.SIGNING_SALT])
    try:
        # Check signature.
        url_protection_string = signer.unsign(token)
        # Check that the given token matches the request parameters.
        random_token = url_protection_string.split(".")[-1]
        if get_qr_url_protection_token(qr_code_options, random_token) != url_protection_string:
            raise PermissionDenied("Request query does not match protection token.")
    except BadSignature:
        raise PermissionDenied("Wrong token signature.")



import base64
import urllib.parse
from collections.abc import Mapping
from datetime import datetime
from typing import Optional, Union, Any

from django.conf import settings
from django.contrib.auth.models import AnonymousUser, User
from django.core.signing import Signer
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.encoding import force_str
from django.utils.safestring import mark_safe
from pydantic import validate_arguments

from qr_code.qrcode import constants, PYDANTIC_CONFIG
from qr_code.qrcode.utils import QRCodeOptions


def _get_default_url_protection_options() -> dict:
    return {
        constants.TOKEN_LENGTH: 20,
        constants.SIGNING_KEY: settings.SECRET_KEY,
        constants.SIGNING_SALT: "qr_code_url_protection_salt",
        constants.ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER: False,
    }


def _get_url_protection_settings() -> Optional[Mapping]:
    if hasattr(settings, "QR_CODE_URL_PROTECTION") and isinstance(settings.QR_CODE_URL_PROTECTION, Mapping):
        return settings.QR_CODE_URL_PROTECTION
    return None


def _options_allow_external_request(url_protection_options: Mapping, user: Union[User, AnonymousUser, None]) -> bool:
    # Evaluate the callable if required.
    if callable(url_protection_options[constants.ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER]):
        allows_external_request = url_protection_options[constants.ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER](user or AnonymousUser())
    elif url_protection_options[constants.ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER] is True:
        allows_external_request = user and user.pk and user.is_authenticated
    else:
        allows_external_request = False
    return allows_external_request


def requires_url_protection_token(user: Union[User, AnonymousUser, None] = None) -> bool:
    return not _options_allow_external_request(get_url_protection_options(), user)


def allows_external_request_from_user(user: Union[User, AnonymousUser, None] = None) -> bool:
    return _options_allow_external_request(get_url_protection_options(), user)


def get_url_protection_options() -> dict:
    options = _get_default_url_protection_options()
    settings_options = _get_url_protection_settings()
    if settings_options is not None:
        options.update(settings.QR_CODE_URL_PROTECTION)
    return options


def _make_random_token() -> str:
    url_protection_options = get_url_protection_options()
    return get_random_string(url_protection_options[constants.TOKEN_LENGTH])


_RANDOM_TOKEN = _make_random_token()


def get_qr_url_protection_signed_token(qr_code_options: QRCodeOptions):
    """Generate a signed token to handle view protection."""
    url_protection_options = get_url_protection_options()
    signer = Signer(key=url_protection_options[constants.SIGNING_KEY], salt=url_protection_options[constants.SIGNING_SALT])
    token = signer.sign(get_qr_url_protection_token(qr_code_options, _RANDOM_TOKEN))
    return token


def get_qr_url_protection_token(qr_code_options, random_token):
    """
    Generate a random token for the QR code image.

    The token contains image attributes so that a user cannot use a token provided somewhere on a website to
    generate bigger QR codes. The random_token part ensures that the signed token is not predictable.
    """
    return ".".join(
        list(
            map(
                str,
                (
                    qr_code_options.size,
                    qr_code_options.border,
                    qr_code_options.version or "",
                    qr_code_options.image_format,
                    qr_code_options.error_correction,
                    random_token,
                ),
            )
        )
    )


def qr_code_etag(request) -> str:
    return f'"{request.path}:{request.GET.urlencode()}:version_{constants.QR_CODE_GENERATION_VERSION_DATE.isoformat()}"'


def qr_code_last_modified(_request) -> datetime:
    return constants.QR_CODE_GENERATION_VERSION_DATE


@validate_arguments(config=PYDANTIC_CONFIG)
def make_qr_code_url(
    data: Any,
    qr_code_options: Optional[QRCodeOptions] = None,
    force_text: bool = True,
    cache_enabled: Optional[bool] = None,
    url_signature_enabled: Optional[bool] = None,
) -> str:
    """Build an URL to a view that handle serving QR code image from the given parameters.

    Any invalid argument related to the size or the format of the image is silently
    converted into the default value for that argument.

    :param str data: Data to encode into a QR code.
    :param QRCodeOptions qr_code_options: The rendering options for the QR code.
    :param bool force_text: Tells whether we want to force the `data` to be considered as text string and encoded in
        byte mode.
    :param bool cache_enabled: Allows to skip caching the QR code (when set to *False*) when caching has
        been enabled.
    :param bool url_signature_enabled: Tells whether the random token for protecting the URL against
        external requests is added to the returned URL. It defaults to *True*.
    """
    qr_code_options = QRCodeOptions() if qr_code_options is None else qr_code_options
    if url_signature_enabled is None:
        url_signature_enabled = constants.DEFAULT_URL_SIGNATURE_ENABLED
    if cache_enabled is None:
        cache_enabled = constants.DEFAULT_CACHE_ENABLED
    cache_enabled_arg = 1 if cache_enabled else 0
    if force_text:
        encoded_data = str(base64.b64encode(force_str(data).encode("utf-8")), encoding="utf-8")
        params = dict(text=encoded_data, cache_enabled=cache_enabled_arg)
    elif isinstance(data, int):
        params = dict(int=data, cache_enabled=cache_enabled_arg)
    else:
        if isinstance(data, str):
            b64data = base64.b64encode(force_str(data).encode("utf-8"))
        else:
            b64data = base64.b64encode(data)
        encoded_data = str(b64data, encoding="utf-8")
        params = dict(bytes=encoded_data, cache_enabled=cache_enabled_arg)
    # Only add non-default values to the params dict
    if qr_code_options.size != constants.DEFAULT_MODULE_SIZE:
        params["size"] = qr_code_options.size
    if qr_code_options.border != constants.DEFAULT_BORDER_SIZE:
        params["border"] = qr_code_options.border
    if qr_code_options.version != constants.DEFAULT_VERSION:
        params["version"] = qr_code_options.version
    if qr_code_options.image_format != constants.DEFAULT_IMAGE_FORMAT:
        params["image_format"] = qr_code_options.image_format
    if qr_code_options.error_correction != constants.DEFAULT_ERROR_CORRECTION:
        params["error_correction"] = qr_code_options.error_correction
    if qr_code_options.micro:
        params["micro"] = 1
    if qr_code_options.eci:
        params["eci"] = 1
    if qr_code_options.boost_error:
        params["boost_error"] = 1
    params["encoding"] = qr_code_options.encoding if qr_code_options.encoding else ""
    params.update(qr_code_options.color_mapping())
    path = reverse("qr_code:serve_qr_code_image")
    if url_signature_enabled:
        # Generate token to handle view protection. The token is added to the query arguments. It does not replace
        # existing plain data query arguments in order to allow usage of the URL as an API (without token since external
        # users cannot generate the signed token!).
        token = get_qr_url_protection_signed_token(qr_code_options)
        params["token"] = token
    url = f"{path}?{urllib.parse.urlencode(params)}"
    return mark_safe(url)



import re

from setuptools import setup

# Get version without importing
with open("qr_code/__init__.py", "rb") as f:
    VERSION = str(re.search('__version__ = "(.+?)"', f.read().decode("utf-8")).group(1))

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("LICENSE", "r", encoding="utf-8") as fh:
    license = fh.read()


setup(
    name="django-qr-code",
    version=VERSION,
    packages=["qr_code", "qr_code.qrcode", "qr_code.templatetags"],
    url="https://github.com/dprog-philippe-docourt/django-qr-code",
    license="BSD 3-clause",
    author="Philippe Docourt",
    author_email="philippe@docourt.ch",
    maintainer="Philippe Docourt",
    description="An application that provides tools for displaying QR codes on your Django site.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["segno>=1.5", "django>=3.2", "pytz", "pydantic"],
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Internet :: WWW/HTTP",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3 :: Only",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Natural Language :: English",
    ],
    keywords="qr code django",
)



from datetime import datetime
from typing import Optional, Union

QR_CODE_GENERATION_VERSION_DATE: datetime = datetime(year=2020, month=9, day=8, hour=12)
SIZE_DICT: dict = {"t": 6, "s": 12, "m": 18, "l": 30, "h": 48}
ERROR_CORRECTION_DICT: dict = {"L": "l", "M": "m", "Q": "q", "H": "h"}
DEFAULT_MODULE_SIZE: Union[str, int] = "m"
DEFAULT_BORDER_SIZE: int = 4
DEFAULT_VERSION: Optional[int] = None
DEFAULT_IMAGE_FORMAT: str = "svg"
DEFAULT_ERROR_CORRECTION: str = "m"
DEFAULT_ECI: bool = False
DEFAULT_BOOST_ERROR: bool = True
DEFAULT_ENCODING: str = "utf-8"
DEFAULT_CACHE_ENABLED: bool = True
DEFAULT_URL_SIGNATURE_ENABLED: bool = True

ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER: str = "ALLOWS_EXTERNAL_REQUESTS_FOR_REGISTERED_USER"
SIGNING_KEY: str = "SIGNING_KEY"
TOKEN_LENGTH: str = "TOKEN_LENGTH"
SIGNING_SALT: str = "SIGNING_SALT"



from django.test import SimpleTestCase

from qr_code.qrcode.maker import make_embedded_qr_code
from qr_code.qrcode.serve import make_qr_code_url
from qr_code.qrcode.utils import QRCodeOptions


class TestIssues(SimpleTestCase):
    def test_reverse_lazy_url(self):
        from django.urls import reverse, reverse_lazy

        options = QRCodeOptions(image_format="svg", size=1)
        url1 = make_qr_code_url(reverse("qr_code:serve_qr_code_image"), options)
        url2 = make_qr_code_url(reverse_lazy("qr_code:serve_qr_code_image"), options)
        self.assertEqual(url1, url2)

        svg1 = make_embedded_qr_code(reverse("qr_code:serve_qr_code_image"), options)
        svg2 = make_embedded_qr_code(reverse_lazy("qr_code:serve_qr_code_image"), options)
        self.assertEqual(svg1, svg2)



"""Tags for Django template system that help generating QR codes."""
from typing import Optional, Any, Union

from django import template

from qr_code.qrcode.maker import make_qr_code_with_args, make_qr_code_url_with_args
from qr_code.qrcode.utils import (
    make_google_play_text,
    make_tel_text,
    make_sms_text,
    make_youtube_text,
    WifiConfig,
    ContactDetail,
    Coordinates,
    EpcData,
    VCard,
    Email,
    MeCard,
    VEvent,
)

register = template.Library()


def _make_app_qr_code_from_obj_or_kwargs(
    obj_or_kwargs, expected_cls, embedded: bool, qr_code_args: dict, extra_qr_code_args: Optional[dict] = None, force_text: bool = True
) -> str:
    if isinstance(obj_or_kwargs, expected_cls):
        obj = obj_or_kwargs
    else:
        # For compatibility with existing views and templates, try to build from dict.
        obj = expected_cls(**obj_or_kwargs)
    final_args = {**qr_code_args}
    if extra_qr_code_args:
        final_args.update(extra_qr_code_args)
    if embedded:
        return make_qr_code_with_args(obj.make_qr_code_data(), qr_code_args=final_args, force_text=force_text)
    else:
        return make_qr_code_url_with_args(obj.make_qr_code_data(), qr_code_args=final_args, force_text=force_text)


def _make_google_maps_qr_code(embedded: bool, **kwargs) -> str:
    if "coordinates" in kwargs:
        coordinates = kwargs.pop("coordinates")
    else:
        coordinates = Coordinates(kwargs.pop("latitude"), kwargs.pop("longitude"))
    if embedded:
        return make_qr_code_with_args(coordinates.make_google_maps_text(), qr_code_args=kwargs)
    else:
        return make_qr_code_url_with_args(coordinates.make_google_maps_text(), qr_code_args=kwargs)


def _make_geolocation_qr_code(embedded: bool, **kwargs) -> str:
    if "coordinates" in kwargs:
        coordinates = kwargs.pop("coordinates")
    else:
        coordinates = Coordinates(kwargs.pop("latitude"), kwargs.pop("longitude"), kwargs.pop("altitude"))
    if embedded:
        return make_qr_code_with_args(coordinates.make_geolocation_text(), qr_code_args=kwargs)
    else:
        return make_qr_code_url_with_args(coordinates.make_geolocation_text(), qr_code_args=kwargs)


@register.simple_tag()
def qr_from_text(text: str, **kwargs) -> str:
    return make_qr_code_with_args(data=text, qr_code_args=kwargs)


@register.simple_tag()
def qr_from_data(data: Any, **kwargs) -> str:
    return make_qr_code_with_args(data=data, qr_code_args=kwargs, force_text=False)


@register.simple_tag()
def qr_for_email(email: Union[str, Email], **kwargs) -> str:
    if isinstance(email, str):
        # Handle simple case where e-mail is simple the electronic address.
        email = Email(to=email)
    return _make_app_qr_code_from_obj_or_kwargs(email, Email, embedded=True, qr_code_args=kwargs)


@register.simple_tag()
def qr_for_tel(phone_number: Any, **kwargs) -> str:
    return make_qr_code_with_args(make_tel_text(phone_number), qr_code_args=kwargs)


@register.simple_tag()
def qr_for_sms(phone_number: Any, **kwargs) -> str:
    return make_qr_code_with_args(make_sms_text(phone_number), qr_code_args=kwargs)


@register.simple_tag()
def qr_for_geolocation(**kwargs) -> str:
    """Accepts a *'coordinates'* keyword argument or a triplet *'latitude'*, *'longitude'*, and *'altitude'*."""
    return _make_geolocation_qr_code(embedded=True, **kwargs)


@register.simple_tag()
def qr_for_google_maps(**kwargs) -> str:
    """Accepts a *'coordinates'* keyword argument or a pair *'latitude'* and *'longitude'*."""
    return _make_google_maps_qr_code(embedded=True, **kwargs)


@register.simple_tag()
def qr_for_youtube(video_id: str, **kwargs) -> str:
    return make_qr_code_with_args(make_youtube_text(video_id), qr_code_args=kwargs)


@register.simple_tag()
def qr_for_google_play(package_id: str, **kwargs) -> str:
    return make_qr_code_with_args(make_google_play_text(package_id), qr_code_args=kwargs)


@register.simple_tag()
def qr_for_contact(contact_detail, **kwargs) -> str:
    return _make_app_qr_code_from_obj_or_kwargs(contact_detail, ContactDetail, embedded=True, qr_code_args=kwargs)


@register.simple_tag()
def qr_for_vcard(vcard, **kwargs) -> str:
    return _make_app_qr_code_from_obj_or_kwargs(vcard, VCard, embedded=True, qr_code_args=kwargs)


@register.simple_tag()
def qr_for_mecard(mecard, **kwargs) -> str:
    return _make_app_qr_code_from_obj_or_kwargs(mecard, MeCard, embedded=True, qr_code_args=kwargs)


@register.simple_tag()
def qr_for_wifi(wifi_config, **kwargs) -> str:
    return _make_app_qr_code_from_obj_or_kwargs(wifi_config, WifiConfig, embedded=True, qr_code_args=kwargs)


@register.simple_tag()
def qr_for_epc(epc_data, **kwargs) -> str:
    extra = dict(
        error_correction="M",
        boost_error=False,
        micro=False,
        encoding="utf-8",
    )
    return _make_app_qr_code_from_obj_or_kwargs(
        epc_data, EpcData, embedded=True, qr_code_args=kwargs, extra_qr_code_args=extra, force_text=False
    )


@register.simple_tag()
def qr_for_event(event, **kwargs) -> str:
    return _make_app_qr_code_from_obj_or_kwargs(event, VEvent, embedded=True, qr_code_args=kwargs)


@register.simple_tag()
def qr_url_from_text(text: str, **kwargs) -> str:
    return make_qr_code_url_with_args(data=text, qr_code_args=kwargs)


@register.simple_tag()
def qr_url_from_data(data: Any, **kwargs) -> str:
    return make_qr_code_url_with_args(data=data, qr_code_args=kwargs, force_text=False)


@register.simple_tag()
def qr_url_for_email(email: Union[str, Email], **kwargs) -> str:
    if isinstance(email, str):
        # Handle simple case where e-mail is simple the electronic address.
        email = Email(to=email)
    return _make_app_qr_code_from_obj_or_kwargs(email, Email, embedded=False, qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_tel(phone_number: Any, **kwargs) -> str:
    return make_qr_code_url_with_args(make_tel_text(phone_number), qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_sms(phone_number: Any, **kwargs) -> str:
    return make_qr_code_url_with_args(make_sms_text(phone_number), qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_geolocation(**kwargs) -> str:
    """Accepts a *'coordinates'* keyword argument or a triplet *'latitude'*, *'longitude'*, and *'altitude'*."""
    return _make_geolocation_qr_code(embedded=False, **kwargs)


@register.simple_tag()
def qr_url_for_google_maps(**kwargs) -> str:
    """Accepts a *'coordinates'* keyword argument or a pair *'latitude'* and *'longitude'*."""
    return _make_google_maps_qr_code(embedded=False, **kwargs)


@register.simple_tag()
def qr_url_for_youtube(video_id: str, **kwargs) -> str:
    return make_qr_code_url_with_args(make_youtube_text(video_id), qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_google_play(package_id: str, **kwargs) -> str:
    return make_qr_code_url_with_args(make_google_play_text(package_id), qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_contact(contact_detail, **kwargs) -> str:
    return _make_app_qr_code_from_obj_or_kwargs(contact_detail, ContactDetail, embedded=False, qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_vcard(vcard, **kwargs) -> str:
    return _make_app_qr_code_from_obj_or_kwargs(vcard, VCard, embedded=False, qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_mecard(mecard, **kwargs) -> str:
    return _make_app_qr_code_from_obj_or_kwargs(mecard, MeCard, embedded=False, qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_wifi(wifi_config, **kwargs) -> str:
    return _make_app_qr_code_from_obj_or_kwargs(wifi_config, WifiConfig, embedded=False, qr_code_args=kwargs)


@register.simple_tag()
def qr_url_for_epc(epc_data, **kwargs) -> str:
    extra = dict(
        error_correction="M",
        boost_error=False,
        micro=False,
        encoding="utf-8",
    )
    return _make_app_qr_code_from_obj_or_kwargs(
        epc_data, EpcData, embedded=False, qr_code_args=kwargs, extra_qr_code_args=extra, force_text=False
    )


@register.simple_tag()
def qr_url_for_event(event, **kwargs) -> str:
    return _make_app_qr_code_from_obj_or_kwargs(event, VEvent, embedded=False, qr_code_args=kwargs)
